/* sharcsolve -- Convert raw SHARC II FITS files, with binary tables, into
   FITS image of sky, using iterative analysis. */

/* Dynamic pointing must be in AZO, ELO, AZ_ERROR, EL_ERROR, RAO, DECO,
   Parallactic Angle, and CHOP_OFFSET columns in SHARC2 Data HDU's */

/* chopping must be along array long axis */

/* Model:

   rawdata[i,t] = gain[i]*{skytran[t]*image[x[i,t],y[i,t]] + background[t]}
                  + offset[i,t] + drift[i,t]

   i = pixel index, t = time index
   offset is independent for different files
   gain is independent for different sets
   drift is on a 'per row' basis and is also independent for different files
   skytran[t] is not a fit parameter
   background[t] has four terms to allow for plane + quadratic in long axis
*/

/* Units:  mV (referred to amplifier output) ABOVE ATMOSPHERE */

/* Future improvements:
   1) Search for correct data HDU instead of absolute move
   2) Divide by amplifier gain to refer signals to input
   3) Use orthogonal form of background quadratic
*/

/* To compile:
       gcc -o sharcsolve sharcsolve.c select.c -L. -lcfitsio -lm -lnsl {-lsocket}
Requires:  fitsio.h, libcfitsio.a, longnam.h */

/* Input file formats:

   calibration file -- not implemented
     line 1 -- log_file scan#_column tau_column
     line 2 -- tau relation 'a'; tau(sharc) = a(tau - b)
     line 3 -- tau relation 'b'
     line 4 -- Jy/mV conversion factor
     lines 5+ -- Jy/mV conversion factor

   fitslist file
     MORE HERE

   mask file
     MORE HERE

   pixel file
     MORE HERE

*/

/* C. Darren Dowell */

/* version 1.0 -- 2002 September 28 -- initial release */
/* version 1.1 -- 2002 November 16 -- 12 rows; 2D HDU 3; saturation checking;
                                      RA/Dec scans; WCS */
/* version 1.2 -- 2002 December 1 -- background gradients */
/* version 1.3 -- 2002 December 2 -- polynomial offsets */
/* version 1.4 -- 2002 December 10 -- row drift term; improved grad. calc. */
/* version 1.5 -- 2002 December 15 -- offset cut */
/* version 1.6 -- 2002 December 15 -- gradient iterations option */
/* version 1.7 -- 2003 January 9 -- radec offset cut; IRC data format */
/* version 1.8 -- 2003 January 10 -- chopped observations */
/* version 2.0 -- 2003 January 23 -- new options (-de, -ma, -r0, -ro, -mu)
                                     some options (-p) obsolete
                                     absolute coordinates in sky map
                                     documentation of solution in HDU 0
                                     pass through copies of input HDU 0's
                                     a little more care with medians
                                     print (-)frame rms even for bad frames */
/* version 2.1 -- 2003 Mar 09 -- noise and signal-to-noise images
                                 redefined -de, -f, and -w options
                                 new options (-ms, -dg)
                                 read ROTZERO from header
                                 fixed some bugs related to scans with
                                   static offsets
                                 updated reduced chi2 calculation
                                 small change in mean_over_array */
/* version 2.2 -- 2003 Mar 28 -- -p option:  pixel rms weighting in map,
                                   background, and rms calculation */
/* version 2.3 -- 2003 Apr 08 -- fixed a few bugs in 2.2, especially
                                   affecting bright sources */
/* version 3.0 -- 2005 Oct  1 -- further timing checking for data
                                   April 2003 and after
                                 -c -> -ch
                                 new option -cf -- NOT IMPLEMENTED
                                 new option -co
                                 background quadratic in x
                                 ROTZERO not required in header
                                   (Nov. 02 data)
                                 -f -> -fa
                                 new option -fi
                                 different way of handling map too small
                                 flag samples with bad elevation as bad
                                 new option -cd
                                 fixed some bugs (forgot to set vars. to 0)
                                 support new format with DSOS HDU
                                 new options -ft, -fl
                                 revised meaning of -ch; new -cl, -ct
                                 support for Nasmyth platform
                                 pointing correct for other reference pixels
                                 first attempt to add nodding offsets
                                 new options -rz, -sc
                                 eliminated -s option
                                 special commands in file list:
                                   $g -- allow different gains for coming set
                                   $t -- different zenith tau
                                 new option -ca
                                 more status info during iterations
                                 new options -sh, -sv
                                 use chop info. in preliminary solution
                                 eliminate azo,elo,rao,deco correlated bgnd.
                                 new option -di
                                 */

/* Usage:  sharcsolve fitslist {pixel.dat, -ac # #, -al, -b, -ca #, -cd, -cf <file>, -ch, -cl #, -co, -ct #, -dc # #, -de #, -dg #, -di #, -du, -ec # #, -er, -fa # #, -fi #, -fl #, -ft #, -g #, -i #, -l #, -ma #, -mo, -ms <file>, -mu #, -n #, -o # #, -p #, -q #, -r0 #, -ra #, -rc # #, -ro #, -rz #, -sc #, -sh, -sv, -t #, -w #, -x, -z #} */
/* -ac # # : azimuth offset max, min */
/* -al : altaz (pointing) mode */
/* -b : block average */
/* -ca # : chop orientation (degrees) */
/* -cd : dual beam chopped image
/* -cf <file> : calibration file */
/* -ch : chopped observation */
/* -cl # : forced chopper throw (arcsec) */
/* -co : write covariance matrix */
/* -ct # : chopper tolerance (arcsec) */
/* -dc # # : Dec. offset max, min */
/* -de # : number of iterations in deep mode (faint, sparse, compact sources) */
/* -dg # : number of sigma for deglitching */
/* -di # : distortion (y scale / x scale) */
/* -du : time record (dump) */
/* -ec # # : elevation offset max, min */
/* -er : residuals image */
/* -fa # # : true FAZO and FZAO */
/* -fi # : minimum frequency to filter */
/* -fl # : clipping level for unfaithful iterations */
/* -ft # : number of faithful iterations */
/* -g # : number of iterations with nonzero atmospheric gradients */
/* -i # : number of iterations with nonzero map */
/* -l # : map size */
/* -ma # : magnification (map pixels per camera pixel) */
/* -mo : movie format */
/* -ms <file> = mask file for initial iteration */
/* -mu # : multiplicative constant */
/* -n # : total number of iterations */
/* -o # # : order of offset and drift functions */
/* -p # : iteration number to recalculate pixel weights */
/* -q # : number of iterations with nonzero atmospheric quadratic */
/* -r0 # : override rotator zero angle */
/* -ra # : avoidance radius in map pixels */
/* -rc # # : RA offset max, min */
/* -ro # : override rotator angle */
/* -rz # : radius in map pixels at which to zero map */
/* -sc # : scale factor (arcsec/bolometer) */
/* -sh # : SHARP H array only */
/* -sv # : SHARP V array only */
/* -t # : zenith opacity */
/* -w # : minimum weight (fraction of peak) */
/* -x : ignore telescope motion -- OBSOLETE? */
/* -z # : additive constant */


#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "fitsio.h"
#include "nrutil.h"

#define  MAX_FILES  100
#define  MAX_O_TERMS  50
#define  B_TERMS  4
#define  MAX_LEN  1000
#define  MAX_LIST  20000
#define  MAX_X  2001
#define  MAX_Y  2001
#define  SCALE_CASS  4.87
#define  SCALE_NAS  4.63
#define  DISTORT_CASS  0.995
#define  DISTORT_NAS  1.018
#define  BAD_MASK  3
#define  ARRAY_COLUMNS  32
#define  ARRAY_ROWS  12
#define  NTILE  50
#define  BLOCK_SIZE  28
#define  NUM_ITER0  10
#define  GRAD_ITER0  2
#define  QUAD_ITER0  0
#define  DEEP_ITER0  5
#define  PWT_ITER0  7
#define  O_ORDER0  1
#define  D_ORDER0  0
#define  DEGL_SIG0  5.0
#define  MUL_CONST0  1.0
#define  PI  3.141592653589
#define  MAX_V  3.5
#define  MIN_V  -3.5
#define  FLUX0  1.0e1
#define  MAX_TIME_JUMP  1.9
#define  MIN_GAIN 2.0e-1
#define  MIN_RMS 3.0e-1
#define  FREQ0  -1.0
#define  NEG_CUTOFF0  0.0
#define  CHOP_TOL0  4.0
#define  CHOP_SCALE  1.0

int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                       float *array);
int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                       double *array);
int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                        char *array);
void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,
               float *chop_lo, float *chop_eff);
void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,
                float chop_tol, char *badsamp, float **rawdata, long ndet,
                float **el, float **ra, float **dec, float **azerr,
                float **elerr, float **azo, float **elo, float **rao,
                float **deco, float **pa, char **scanning);
void sign_correct(float *rawdata, float offset[], long ndet, long nsamp);
void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,
                      char *badsamp);
void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,
                  float sampdur);
void offset_cut(long nsamp, float *azo, float *azerr, float *elo,
                float *elerr, float *el, float sampdur, char *badsamp,
                float  azo0, float elo0, float ao_max, float ao_min,
                float eo_max, float eo_min, float *rao, float *deco,
                float ro_max, float ro_min, float do_max, float do_min,
                int year);
void pixel_ntile_over_file(float *rawdata, float *offset, long ndet,
                           long nsamp, char *badsamp);
void pixel_mean_over_file(float *rawdata, float gain[], float offset[],
                          float *background, long ndet, long nsamp,
                          char *badsamp, float *x0, float *y0, float *x1,
                          float *y1, float avoid_r2, char *mask,
                          char chop_flag, int chop_dx, int chop_dy);
void ntile_over_array(float *rawdata, float gain[], float *offset,
                      float *background, long ndet, long nsamp, int baddet[]);
void mean_over_array(float *rawdata, float gain[], float offset[],
                     float *background, long ndet, long nsamp, char *badsamp,
                     int baddet[], float *x0, float *y0, float *x1, float *y1,
                     float avoid_r2, char *mask, char chop_flag, int chop_dx,
                     int chop_dy);
void pre_filter(float *rawdata, float gain[], float offset[],
                float *background, long ndet, long nsamp, char *badsamp,
                int baddet[], float *x0, float *y0, float *x1, float *y1,
                float avoid_r2, char *mask, float filter_freq, float sampdur);
void calc_xy_vs_time(float *x0, float *y0, long nsamp, float *el, float *ra,
                     float *dec, float *azo, float *elo, float *azerr,
                     float *elerr, float *rao, float *deco, float *pa,
                     char fixed_flag, float true_fazo, float true_fzao,
                     float fazo, float fzao, char altaz_flag, float rao0,
                     float deco0, float azo0, float elo0);
void calc_pixel_xy(long det, float *x, float *y);
void calc_rot_pixel_xy(float *x1, float *y1, long ndet, long nsamp, float *pa,
                       float rotator, float *el, char altaz_flag,
                       char nasmyth_flag);
void calc_chop_dx(int nfiles, long nsamp[], float *pa[], float rotator[],
                  char altaz_flag, float chop_throw, char nasmyth_flag,
                  int *chop_dx, int *chop_dy, float chop_angle);
void chop_combine(float *image, float *weight, float *im_rms, int dx, int dy);
void get_xy(long pix, long samp, float *x0, float *y0, float *x1, float *y1,
            int *x, int *y, char bad_flag);
void calc_basfunc(float *basfunc[], int order, long nsamp, char *badsamp,
                  char *scanning, char scanning_flag);
void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],
               float tauzen);
float get_offset(float *offset[], float *basfunc[], int o_order, long nsamp,
                 long samp, long det);
float get_drift(float *drift[], float *basfunc[], int o_order, int d_order,
                long nsamp, long samp, long det, char deep_flag,
                float *rowavg[]);
float get_background(float *background[], long samp, long det);
void calc_map(float *image, float *weight, float *skytran[], float *rawdata[],
              float *gain[], int gset[], float p_wt[], int baddet[],
              char *badsamp[], float *offset[], float *basfunc[], int o_order,
              float *background[], float *drift[], int d_order, float *x0[],
              float *y0[], float *x1[], float *y1[], int nfiles, long nsamp[],
              long ndet, char *scanning[], char scanning_flag, char deep_flag,
              float *rowavg[], float *mean, char rz_flag, float zero_radius,
              char chop_flag, int chop_dx, int chop_dy);
void remove_neg(float *image, float neg_cutoff);
float calc_rms(float *image, float *weight, float *im_rms, float *skytran[],
              float *rawdata[], float *gain[], int gset[], float p_wt[],
              int baddet[], char *badsamp[], float *offset[], float *basfunc[],
              int o_order, float *background[], float *f_wt[], float *drift[],
              int d_order, float *x0[], float *y0[], float *x1[], float *y1[],
              int nfiles, long nsamp[], long ndet, char *scanning[],
              char scanning_flag, char deep_flag, float *rowavg[], float mean,
              char calc_im_rms, char calc_p_rms, char calc_f_rms);
void deglitch_map(float *image, float *im_rms, float *weight, float *skytran[],
                  float *rawdata[], float *gain[], int gset[], float p_wt[],
                  int baddet[], char *badsamp[], float *offset[],
                  float *basfunc[], int o_order, float *background[],
                  float *drift[], int d_order, float *x0[], float *y0[],
                  float *x1[], float *y1[], int nfiles, long nsamp[],
                  long ndet, char *scanning[], char scanning_flag,
                  char deep_flag, float *rowavg[], float *mean, char rz_flag,
                  float zero_radius, char chop_flag, int chop_dx, int chop_dy,
                  float deg_thresh, long *deg_count, char drop_list[]);
void calc_offsets(float *image, float *weight, float *skytran[],
                  float *rawdata[], float *gain[], int gset[], int baddet[],
                  char *badsamp[], float *offset[], float *basfunc[],
                  int o_order, float *background[], float *drift[],
                  int d_order, float *x0[], float *y0[], float *x1[],
                  float *y1[], int nfiles, long nsamp[], long ndet,
                  char *scanning[], char scanning_flag, char deep_flag,
                  float *rowavg[]);
void calc_drift(float *image, float *weight, float *skytran[],
                float *rawdata[], float *gain[], int gset[], int baddet[],
                char *badsamp[], float *offset[], float *basfunc[], int o_order,
                float *background[], float *drift[], int d_order, float *x0[],
                float *y0[], float *x1[], float *y1[], int nfiles,
                long nsamp[], long ndet, char *scanning[], char scanning_flag,
                float *rowavg[]);
void calc_gains(float *image, float *weight, float *skytran[], float *rawdata[],
                float *gain[], int gset[], int ngainsets, int baddet[],
                char *badsamp[], float *offset[], float *basfunc[], int o_order,
                float *background[], float *drift[], int d_order, float *x0[],
                float *y0[], float *x1[], float *y1[], int nfiles,
                long nsamp[], long ndet, char *scanning[], char scanning_flag,
                char deep_flag, float *rowavg[]);
void norm_gains(float *gain[], int ngainsets, int baddet[], long ndet);
float avg_p_wt(float p_wt[], int baddet[], long ndet);
void calc_background(float *image, float *weight, float *skytran[],
                     float *rawdata[], float *gain[], int gset[], float p_wt[],
                     int baddet[], char *badsamp[], float *offset[],
                     float *basfunc[], int o_order, float *background[],
                     float *drift[], int d_order, float *x0[], float *y0[],
                     float *x1[], float *y1[], int nfiles, long nsamp[],
                     long ndet, char *scanning[], char scanning_flag,
                     char gradient_flag, char quadratic_flag, char deep_flag,
                     float *rowavg[]);
void zero_background(float *background[], int nfiles, long nsamp[],
                     char *badsamp[], float *azo[], float *azerr[], float *el[],
                     float *elo[], float *elerr[], float *rao[], float *deco[]);
void block_average(float *rawdata, long ndet, long *nsamp);
void precess2000(float ra_1, float dec_1, float eq_1, float *ra_0,
                 float *dec_0);
void printabort(char *message);
void printerror(int status);
float element_select(unsigned long k, unsigned long n, float arr[]);
void gaussj(float **a, int n, float **b, int m);
void realft(float data[], unsigned long n, int isign);

char usage[] = "Usage:  sharcsolve fitslist {pixel.dat,-ac # #,-al,-b,-ca,-cd,-cf <file>,-ch,-cl #,-co,-ct #,-dc # #,-de #,-dg #,-di #,-du,-ec # #,-er,-fa # #,-fi #,-fl #,-ft #,-g #,-i #,-l #,-ma #,-mo,-ms <file>,-mu #,-n #,-o # #,-p #,-q #,-r0 #,-ra #, -rc # #,-ro #,-rz #,-sc #,-sh,-sv,-t #,-w #,-x,-z #}\n    -ac # # = azimuth offset max, min\n    -al = altaz (pointing) mode\n    -b = block average\n    -ca # = chopper orientation (degrees)\n    -cd = dual beam chopped image\n    -cf <file> = calibration file\n    -ch = chopped observation\n    -cl # = forced chopper throw (arcsec)\n    -co = write covariance matrix\n    -ct # = chopper tolerance (arcsec)\n    -dc # # = Dec. offset max, min\n    -de # = number of iterations in deep mode (faint, sparse, compact sources)\n    -dg # = number of sigma for deglitching\n    -di # = distortion (y scale vs x scale)\n    -du = time record (dump)\n    -ec # # = elevation offset max, min\n    -er = residuals image\n    -fa # # = true FAZO and FZAO\n    -fi # = minimum frequency to filter\n    -fl # = clipping level for unfaithful iterations\n    -ft # = number of faithful iterations\n    -g # = number of iterations with nonzero atmospheric gradients\n    -i # = number of iterations with nonzero map\n    -l # = map size\n    -ma # = magnification (map pixels per camera pixel)\n    -mo = movie format\n    -ms <file> = mask file for initial iteration\n    -mu # = multiplicative constant\n    -n # = total number of iterations\n    -o # # = order of offset and drift functions\n    -p # = iteration number to recalculate pixel weights\n    -q # = number of iterations with nonzero atmospheric quadratic\n    -r0 # = override rotator zero angle\n    -ra # = avoidance radius in map pixels\n    -rc # # = RA offset max, min\n    -ro # = override rotator angle\n    -rz # = radius in map pixels at which to zero map\n    -sc # = scale factor(arcsec/bolometer)\n    -sh = SHARP H array\n    -sv = SHARP V array\n    -t # = zenith opacity\n    -w # = minimum weight (fraction of peak)\n    -x = ignore telescope motion\n    -z # = additive constant\n";
char version[] = "sharcsolve 3.0, 2005 Oct  1  IN DEVELOPMENT";

/* make global variables to save programming time: */
int  IM_X=201, IM_Y=201;
float  IM_X0=101.0, IM_Y0=101.0;
float  ROTATOR_0 = 16.0;
float  MAGNIFICATION = 2.0;    /* map pixels/detector */
char  map_too_small = 'n';
float  A_CEN_X, A_CEN_Y;
float  SCALE = SCALE_CASS;     /* arcsec/detector */
float  DISTORT = DISTORT_CASS;

int main(int argc, char *argv[])

{
    float  *gain[MAX_FILES], g;     /* indices 0 to ngainsets-1, 1 to ndet */
    int  gset[MAX_FILES];     /* index 0 to ngainsets-1 */
    float  p_wt[ARRAY_ROWS*ARRAY_COLUMNS+1], pw;     /* index 1 to ndet */
    int  baddet[ARRAY_ROWS*ARRAY_COLUMNS+1], bd;     /* index 1 to ndet */
    float  *offset[MAX_FILES*MAX_O_TERMS], o;
                 /* indices 0 to nfiles-1, 0 to o_order, 0 to ndet-1 */
    float  *drift[MAX_FILES*MAX_O_TERMS], d;
                 /* indices 0 to nfiles-1, 0 to d_order-1, 0 to ARRAY_ROWS-1 */
    float  *basfunc[MAX_FILES*MAX_O_TERMS];
                 /* indices 0 to nfiles-1, 0 to nterms-1, 0 to nsamp-1 */
    float  *background[MAX_FILES*B_TERMS], *f_wt[MAX_FILES], b;
                 /* indices 0 to nfiles-1, 0 to B_TERMS-1, 0 to nsamp-1 */
    float  *rowavg[MAX_FILES*ARRAY_ROWS];
                 /* indices 0 to nfiles-1, 0 to ARRAY_ROWS-1, 0 to nsamp-1 */
    char  *badsamp[MAX_FILES];     /* indices 0 to nfiles-1, 0 to nsamp-1 */
    int  i, j;
    int  arg;
    char  output_flag, arg_flag, avg_flag, fixed_flag, scanning_flag, first;
    char  residual_flag, ignore_flag, altaz_flag, gradient_flag, chop_flag;
    char  ro_flag, r0_flag, deep_flag, mask_flag, pwt_flag, cov_flag, rz_flag;
    char  quadratic_flag, chop_dual_flag, nasmyth_flag, sharp_flag;
    FILE  *infile, *outfile;
    char  line[MAX_LEN], headline[MAX_LEN], filename[MAX_LEN], source[MAX_LEN];
    char  date[MAX_LEN], maskfilename[MAX_LEN], mask_type[MAX_LEN];
    char  headline2[MAX_LEN], platform[MAX_LEN];
    int  nfiles, ngainsets;
    char  newset;
    fitsfile  *infptr, *outfptr;
    int  status = 0;     /* must be initialized to zero! */
    float  rao0[MAX_FILES], deco0[MAX_FILES];
    float  azo0[MAX_FILES], zao0[MAX_FILES];
    float  fazo[MAX_FILES], fzao[MAX_FILES], rotator[MAX_FILES];
    float  azo_chop[MAX_FILES];
    float  true_fazo[MAX_FILES], true_fazo0, true_fzao[MAX_FILES], true_fzao0;
    float  rotator_override, rotator0_override, rotator0;
    int  hdunum, firstdatahdu, hdutype;
    long  nsamp[MAX_FILES], ndet, ncol0, nrow, row;
    long  nelements;
    float  *rawdata[MAX_FILES];     /* indices 0 to nfiles-1, 0 to nsamp-1,
                                       0 to ndet-1 */

    /* indices 0 to nfiles-1, 0 to nsamp-1 : */
    float  *el[MAX_FILES], *ra[MAX_FILES], *dec[MAX_FILES];
    float  *azerr[MAX_FILES], *elerr[MAX_FILES];
    float  *azo[MAX_FILES], *elo[MAX_FILES];
    float  *rao[MAX_FILES], *deco[MAX_FILES], *pa[MAX_FILES], *chop[MAX_FILES];
    char  *scanning[MAX_FILES], *on_beam[MAX_FILES], *off_beam[MAX_FILES];
    double  *ut, *det_time;
    float  *x0[MAX_FILES], *y0[MAX_FILES], *x1[MAX_FILES], *y1[MAX_FILES];
    float  *skytran[MAX_FILES];

    int  file, set;
    long  samp, firstsamp, det, det1, det2;
    int  x, y;
    float  *image, *weight, *im_rms;
    char  *mask;
    int  bitpix, naxis;
    long  naxes[3], fpixel[3];
    int  iter;
    double  sum0, sum1, sum2, sum3, sum4;
    float  mean;
    int  num_iter, map_iter, grad_iter, quad_iter, deep_iter, pwt_iter;
    int  faithful_iter;
    float  avoid_r2;
    float  weight_cut;
    float  zero, mul_const;
    int  o_order, d_order, term;
    float  tauzen[MAX_FILES], tauzen0;
    float  rms;
    float  raep, decep, equinox;
    float  val;
    int  ival;
    int  worst_pixel;
    float  worst_pixel_rms;
    int  worst_frame, worst_frame_file;
    float  worst_frame_rms;
    float  max_weight;
    double  good_det, good_samp, fit_param;
    float  sampdur;
    float  ao_max = 1.0e30, ao_min = -1.0e30;
    float  eo_max = 1.0e30, eo_min = -1.0e30;
    float  ro_max = 1.0e30, ro_min = -1.0e30;
    float  do_max = 1.0e30, do_min = -1.0e30;
    int  colnum, typecode;
    long  width;
    float  chop_throw = -1.0, chop_tol = CHOP_TOL0;
    float  chop_hi, chop_lo, chop_eff;
    int  chop_dx = 0, chop_dy = 0;
    float  chop_angle = 0.0;
    long  deg_count;
    int  x2, x3, y2, y3;
    int  year;
    float  deglitch_sigma;
    time_t  the_time;
    float  *covar;
    float  filter_freq = FREQ0;
    float  old_nsamp, coadd_factor;
    float  neg_cutoff = NEG_CUTOFF0;
    float  zero_radius;
    float  scale_override = -1.0, distort_override = -1.0;
    char  drop_list[MAX_FILES], drop = 0;

    /************* READ COMMAND LINE ***************/

    fprintf(stderr, "%s\n", version);
    if (argc < 2)
      printabort(usage);
    for (j = 0; j < MAX_FILES; j++)
      gain[j] = (float *) malloc((ARRAY_ROWS*ARRAY_COLUMNS+1)*sizeof(float));
    for (i = 1; i <= ARRAY_ROWS*ARRAY_COLUMNS; i++) {
      baddet[i] = 0;
      p_wt[i] = 1.0;
      for (j = 0; j < MAX_FILES; j++)
        gain[j][i] = 1.0;
    }
    arg = 2;
    output_flag = 's';
    avg_flag = 'n';
    fixed_flag = 'n';
    scanning_flag = 1;     /* no longer any options to change this */
    num_iter = NUM_ITER0;
    map_iter = -9999;
    faithful_iter = -9999;
    grad_iter = GRAD_ITER0;
    quad_iter = QUAD_ITER0;
    deep_iter = DEEP_ITER0;
    pwt_iter = PWT_ITER0;
    avoid_r2 = -1.0;
    weight_cut = -1.0;
    zero = 0.0;
    mul_const = MUL_CONST0;
    residual_flag = 'n';
    ignore_flag = 'n';
    altaz_flag = 'n';
    o_order = O_ORDER0;
    d_order = D_ORDER0;
    tauzen0 = 0.0;
    chop_flag = 'n';
    chop_dual_flag = 'n';
    ro_flag = 'n';
    rz_flag = 'n';
    r0_flag = 'n';
    mask_flag = 'n';
    cov_flag = 'n';
    sharp_flag = 'n';
    deglitch_sigma = DEGL_SIG0;
    while (arg < argc) {
      arg_flag = 'n';
      if (strncmp(argv[arg], "-ac", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &ao_max) != 1)
          printabort("Bad azimuth offset cut");
        arg++;
        if (sscanf(argv[arg], "%f", &ao_min) != 1)
          printabort("Bad azimuth offset cut");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-al", 3) == 0) {
        altaz_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-b", 2) == 0) {
        avg_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ca", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &chop_angle) != 1)
          printabort("Bad chop angle");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-cd", 3) == 0) {
        chop_dual_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ch", 3) == 0) {
        chop_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-cl", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &chop_throw) != 1)
          printabort("Bad chopper throw");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-co", 3) == 0) {
        cov_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ct", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &chop_tol) != 1)
          printabort("Bad chopper tolerance");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-dc", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &do_max) != 1)
          printabort("Bad Dec. offset cut");
        arg++;
        if (sscanf(argv[arg], "%f", &do_min) != 1)
          printabort("Bad Dec. offset cut");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-de", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &deep_iter) != 1)
          printabort("Bad number of deep iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-dg", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &deglitch_sigma) != 1)
          printabort("Bad sigma for deglitching");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-di", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &distort_override) != 1)
          printabort("Bad distortion");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-du", 3) == 0) {
        output_flag = 'd';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ec", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &eo_max) != 1)
          printabort("Bad elevation offset cut");
        arg++;
        if (sscanf(argv[arg], "%f", &eo_min) != 1)
          printabort("Bad elevation offset cut");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-er", 3) == 0) {
        residual_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-fa", 3) == 0) {
        fixed_flag = 'y';
        arg++;
        if (sscanf(argv[arg], "%f", &true_fazo0) != 1)
          printabort("Bad true FAZO");
        arg++;
        if (sscanf(argv[arg], "%f", &true_fzao0) != 1)
          printabort("Bad true FZAO");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-fi", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &filter_freq) != 1)
          printabort("Bad filter frequency");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-fl", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &neg_cutoff) != 1)
          printabort("Bad negative cutoff level");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ft", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &faithful_iter) != 1)
          printabort("Bad number of faithful iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-g", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &grad_iter) != 1)
          printabort("Bad number of gradient iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-i", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &map_iter) != 1)
          printabort("Bad number of map iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-l", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &IM_X) != 1)
          printabort("Bad map size");
        if ((IM_X > MAX_X) || (IM_X > MAX_Y))
          printabort("Map size too large");
        IM_Y = IM_X;
        IM_X0 = (IM_X + 1.0)/2.0;
        IM_Y0 = IM_X0;
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ma", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &MAGNIFICATION) != 1)
          printabort("Bad magnification");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-mo", 3) == 0) {
        output_flag = 'm';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ms", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%s", maskfilename) != 1)
          printabort("Bad mask file name");
        mask_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-mu", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &mul_const) != 1)
          printabort("Bad multiplicative constant");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-n", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &num_iter) != 1)
          printabort("Bad number of total iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-o", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &o_order) != 1)
          printabort("Bad offset order");
        if (o_order >= MAX_O_TERMS)
          printabort("Offset order too large");
        arg++;
        if (sscanf(argv[arg], "%d", &d_order) != 1)
          printabort("Bad drift order");
        if ((o_order+d_order) >= MAX_O_TERMS)
          printabort("Drift order too large");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-p", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &pwt_iter) != 1)
          printabort("Bad entry for pixel weighting iteration");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-q", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%d", &quad_iter) != 1)
          printabort("Bad number of quadratic iterations");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-r0", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &rotator0_override) != 1)
          printabort("Bad rotator zero angle");
        r0_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ra", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &avoid_r2) != 1)
          printabort("Bad avoidance radius");
        avoid_r2 = avoid_r2*avoid_r2;
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-rc", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &ro_max) != 1)
          printabort("Bad RA offset cut");
        arg++;
        if (sscanf(argv[arg], "%f", &ro_min) != 1)
          printabort("Bad RA offset cut");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-ro", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &rotator_override) != 1)
          printabort("Bad rotation angle");
        ro_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-rz", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &zero_radius) != 1)
          printabort("Bad zero radius");
        rz_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-sc", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &scale_override) != 1)
          printabort("Bad scale");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-sh", 3) == 0) {
        sharp_flag = 'h';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-sv", 3) == 0) {
        sharp_flag = 'v';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-t", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &tauzen0) != 1)
          printabort("Bad zenith tau");
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-w", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &weight_cut) != 1)
          printabort("Bad minimum weight");
        if (weight_cut >= 1.0) {
          weight_cut = 0.25;
          fprintf(stderr, "  WARNING:  Bad minimum weight, setting to %4.2f",
                  weight_cut);
          system("date >> sharcsolve_errlog.txt");
          system(
            "echo \"WARNING:  Bad minimum weight\" >> sharcsolve_errlog.txt");
        }
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-x", 2) == 0) {
        ignore_flag = 'y';
        arg_flag = 'y';
      }
      if (strncmp(argv[arg], "-z", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &zero) != 1)
          printabort("Bad additive constant");
        arg_flag = 'y';
      }
      if (arg_flag == 'n') {
        if ((infile = fopen(argv[arg], "r")) == NULL) {
          sprintf(line, "Didn't understand %s", argv[arg]);
          printabort(line);
        }
        while (fgets(line, MAX_LEN, infile) != NULL) {
          if (sscanf(line, "%d %d %f %d %f", &y, &x, &g, &bd, &pw) == 5) {
            det = (y - 1)*ARRAY_COLUMNS + x;
            if ((det >= 1) && (det <= ARRAY_ROWS*ARRAY_COLUMNS)) {
              baddet[det] = bd;
              if ( (fabs(g) < MIN_GAIN)
                   && ((baddet[det] & BAD_MASK) == 0) ) {
                g = MIN_GAIN;
                fprintf(stderr, "WARNING:  Unflagged pixel with low gain\n");
                system("date >> sharcsolve_errlog.txt");
                system(
  "echo \"WARNING:  Unflagged pixel with low gain\" >> sharcsolve_errlog.txt");
              }
              for (set = 0; set < MAX_FILES; set++)
                gain[set][det] = g;
              if (fabs(pw) < MIN_RMS) {
                fprintf(stderr, "WARNING:  Pixel with low rms\n");
                system("date >> sharcsolve_errlog.txt");
                system(
  "echo \"WARNING:  Pixel with low rms\" >> sharcsolve_errlog.txt");
                pw = MIN_RMS;
              }
              p_wt[det] = 1.0/(pw*pw);
            }
          }
        }
        fclose(infile);
      }
      arg++;
    }
    if (map_iter < -9998)
      map_iter = num_iter;
    if (faithful_iter < -9998)
      faithful_iter = num_iter;
    norm_gains(gain, ngainsets, baddet, (long) (ARRAY_COLUMNS*ARRAY_ROWS));
    mask = (char *) malloc(IM_X*IM_Y*sizeof(char));
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        mask[(y-1)*IM_X + (x-1)] = 0;
      }
    }
    if (mask_flag == 'y') {
      if ((infile = fopen(maskfilename, "r")) == NULL) {
        sprintf(line, "Couldn't open mask file %s", maskfilename);
        printabort(line);
      }
      while (fgets(line, MAX_LEN, infile) != NULL) {
        if (sscanf(line, "%s", &mask_type) == 1) {
          switch (mask_type[0]) {
            case 'r': case 'R':
              if (sscanf(line, "%*s %d %d %d %d", &x2, &y2, &x3, &y3) != 4)
                printabort("Didn't understand masking rectangle command");
              for (x = x2; x <= x3; x++) {
                for (y = y2; y <= y3; y++) {
                  if ((x >= 1) && (x <= IM_X) && (y >= 1) && (y <= IM_Y))
                    mask[(y-1)*IM_X + (x-1)] = 1;
                }
              }
              break;
            case '!':
              break;
            default:
              printabort("Didn't understand masking command");
          }
        }
      }
      fclose(infile);
    }

    /************* READ RAW DATA ***************/

    fprintf(stderr, "Reading raw data...\n");
    if ((infile = fopen(argv[1], "r")) == NULL)
      printabort("Could not open list of FITS files");
    nfiles = 0;
    ngainsets = 0;
    newset = 'y';

    /* loop through input file */
    while (fgets(line, MAX_LEN, infile) != NULL) {

     if (line[0] == '$') {
      /* parse special command */
      if ((line[1] == 'g') || (line[1] == 'G')) {
        if (newset != 'y') {
          ngainsets++;
          newset = 'y';
        }
      }
      if ((line[1] == 't') || (line[1] == 'T')) {
        if (sscanf(&(line[2]), "%f", &tauzen0) != 1)
          printabort("Bad zenith tau in file list");
      }
      if ((line[1] == 'd') || (line[1] == 'D'))
        drop = 1;
      if ((line[1] == 'k') || (line[1] == 'K'))
        drop = 0;
      if ((line[1] == 'f') || (line[1] == 'F')) {
        if (sscanf(&(line[2]), "%f %f", &true_fazo0, &true_fzao0) != 1)
          printabort("Bad FAZO FZAO in file list");
        fixed_flag = 'y';
      }
      /* WORK HERE -- check for other special commands */
     }

     else {
      /* read raw FITS file */
      sscanf(line, "%s", filename);     /* Has terminating 0 */
      if (fits_open_file(&infptr, filename, READONLY, &status))
        printerror(status);
      gset[nfiles] = ngainsets;
      newset = 'n';
      /* Read HDU 0 */
      if (fits_read_key(infptr, TSTRING, "OBJECT", source, headline, &status))
        printerror(status);
      fprintf(stderr, "File %d:  %s, source = %s\n", nfiles+1, filename,
              source);
      nasmyth_flag = 'n';
      if (fits_read_key(infptr, TSTRING, "PLATFORM", platform, headline,
                        &status)) {
        status = 0;     /* pre-Aug. '04 data doesn't have PLATFORM in header */
      }
      else {
        if (strncmp(platform, "NA", 2) == 0)
          nasmyth_flag = 'y';
          fprintf(stderr, "  Nasmyth platform detected\n");
          SCALE = SCALE_NAS;       /* only one SCALE allowed per reduction */
          DISTORT = DISTORT_NAS;   /* only one DISTORT allowed per reduction */
      }
      if (scale_override > 0.0)
        SCALE = scale_override;
      if (distort_override > 0.0)
        DISTORT = distort_override;
      if (fits_read_key(infptr, TSTRING, "DATE-OBS", date, headline, &status))
        printerror(status);
      if (sscanf(date, "%d", &year) != 1)
        printabort("Could not read date from FITS file");
      if (fits_read_key(infptr, TFLOAT, "RAEP", &raep, headline, &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "DECEP", &decep, headline, &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "EQUINOX", &equinox, headline, &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "RAO", &rao0[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "DECO", &deco0[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "AZO", &azo0[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "ZAO", &zao0[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "FAZO", &fazo[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "FZAO", &fzao[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "ROTATOR", &rotator[nfiles], headline,
                        &status))
        printerror(status);
      if (fits_read_key(infptr, TFLOAT, "AZO_CHOP", &azo_chop[nfiles],
                        headline, &status)) {
        printerror(status);
        azo_chop[nfiles] = 0.0;
      }
      if (nasmyth_flag == 'y')
        rotator[nfiles] = 0.0;
      if (ro_flag == 'y')
        rotator[nfiles] = rotator_override;
      if (fits_read_key(infptr, TFLOAT, "ROTZERO", &rotator0, headline,
                        &status)) {   /* Seems to handle string keyword OK */
        status = 0;     /* November 2002 data doesn't have ROTZERO in header */
      }
      else {
        ROTATOR_0 = rotator0;
      }
      if (r0_flag == 'y')
        ROTATOR_0 = rotator0_override;
      fprintf(stderr, "  Using ROTATOR_0 = %6.1f\n", ROTATOR_0);
      if ((rotator[nfiles] < -360.0) || (rotator[nfiles] > 360.0)) {
        rotator[nfiles] = ROTATOR_0;
        fprintf(stderr, "  WARNING:  Bad rotator value, setting to ");
        system("date >> sharcsolve_errlog.txt");
        system("echo \"WARNING:  Bad rotator value\" >> sharcsolve_errlog.txt");
      }
      rotator[nfiles] = ROTATOR_0-rotator[nfiles];
      fprintf(stderr, "  Cal. inst. rot. = %6.1f (file %d)\n", rotator[nfiles],
              nfiles+1);

      /* Read data HDU(s) */
      /* find first data HDU */
      hdunum = 5;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) != 0)
        printerror(status);
      if (fits_read_key(infptr, TSTRING, "EXTNAME", headline, headline2,
                        &status))
        printerror(status);
      firstdatahdu = 5;
      if (strncmp(headline, "SHARC2 Data", 11) != 0)
        firstdatahdu = 6;     /* hdunum = 5 is presumably CSO DSOS */

      /* count total number of frames */
      hdunum = firstdatahdu;
      row = 0;
      while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) {
        if (hdutype != BINARY_TBL)
          printabort("Data HDU of wrong type");
        if (fits_get_num_rows(infptr, &nrow, &status))
          printerror(status);
        row += nrow;
        hdunum++;
      }
      nsamp[nfiles] = row;
      status = 0;

      /* allocate memory */
      hdunum = firstdatahdu;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status));
        printerror(status);
      if (hdutype != BINARY_TBL)
        printabort("Data HDU of wrong type");
      if (fits_get_colnum(infptr, CASEINSEN, "Pixel Readout", &colnum, &status))
        printerror(status);
      if (fits_get_coltype(infptr, colnum, &typecode, &ndet, &width, &status))
        printerror(status);
      if (ndet != (ARRAY_ROWS*ARRAY_COLUMNS))
        printabort("Unexpected number of detectors");
      rawdata[nfiles] = (float *) malloc(nsamp[nfiles]*ndet*sizeof(float));
      el[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      ra[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      dec[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      azerr[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      elerr[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      azo[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      elo[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      rao[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      deco[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      pa[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      scanning[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      on_beam[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      off_beam[nfiles] = (char *) malloc(nsamp[nfiles]*sizeof(char));
      chop[nfiles] = (float *) malloc(nsamp[nfiles]*sizeof(float));
      ut = (double *) malloc(nsamp[nfiles]*sizeof(double));
      det_time = (double *) malloc(nsamp[nfiles]*sizeof(double));
      if (fits_read_key(infptr, TFLOAT, "CRPIX3", &A_CEN_Y, headline,
                        &status)) {
        /* This is only for JSharc data, I believe */
        status = 0;
        A_CEN_X = (ARRAY_COLUMNS - 1.0)/2.0;
        A_CEN_Y = (ARRAY_ROWS - 1.0)/2.0;
      }
      else {
        A_CEN_Y -= 1.0;
        if (fits_read_key(infptr, TFLOAT, "CRPIX2", &A_CEN_X, headline,
                          &status))
          printabort("Trouble reading CRPIX2");
        A_CEN_X -= 1.0;
      }
      if (sharp_flag == 'h') {
        fprintf(stderr, "  SHARP H array\n");
        A_CEN_X = 5.5;
        A_CEN_Y = 5.5;
        for (i = 0; i < ARRAY_ROWS; i++)
          for (j = 12; j < ARRAY_COLUMNS; j++)
            baddet[i*ARRAY_COLUMNS+j+1] = 1;
      }
      if (sharp_flag == 'v') {
        fprintf(stderr, "  SHARP V array\n");
        A_CEN_X = 25.5;
        A_CEN_Y = 5.5;
        for (i = 0; i < ARRAY_ROWS; i++)
          for (j = 0; j < (ARRAY_COLUMNS-12); j++)
            baddet[i*ARRAY_COLUMNS+j+1] = 1;
      }
      fprintf(stderr, "  Reference pixel:  %.1f, %.1f\n", A_CEN_X+1,
              A_CEN_Y+1);

      /* read data */
      hdunum = firstdatahdu;
      row = 0;
      while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) {
        if (hdutype != BINARY_TBL)
          printabort("Data HDU of wrong type");
        if (fits_get_num_rows(infptr, &nrow, &status))
          printerror(status);
        if (read_float_column(infptr, "Pixel Readout", nrow, &ncol0,
                              &(rawdata[nfiles][row*ndet])))
          printerror(status);
        if (read_float_column(infptr, "EL", nrow, &ncol0, &(el[nfiles][row])))
          printerror(status);
        if (read_float_column(infptr, "RA", nrow, &ncol0, &(ra[nfiles][row])))
          printerror(status);
        if (read_float_column(infptr, "DEC", nrow, &ncol0,
                              &(dec[nfiles][row])))
          printerror(status);
        if (read_float_column(infptr, "AZ_ERROR", nrow, &ncol0,
                              &(azerr[nfiles][row])))
          printerror(status);
        if (read_float_column(infptr, "EL_ERROR", nrow, &ncol0,
                              &(elerr[nfiles][row])))
          printerror(status);
        if (read_float_column(infptr, "AZO", nrow, &ncol0,
                              &(azo[nfiles][row])))
          if (read_float_column(infptr, "AZ_OFFSET", nrow, &ncol0,
                                &(azo[nfiles][row])))
            printerror(status);
        if (read_float_column(infptr, "ELO", nrow, &ncol0,
                              &(elo[nfiles][row])))
          if (read_float_column(infptr, "EL_OFFSET", nrow, &ncol0,
                                &(elo[nfiles][row])))
            printerror(status);
        if (read_float_column(infptr, "RAO", nrow, &ncol0,
                              &(rao[nfiles][row])))
          if (read_float_column(infptr, "RA_OFFSET", nrow, &ncol0,
                                &(rao[nfiles][row])))
            printerror(status);
        if (read_float_column(infptr, "DECO", nrow, &ncol0,
                              &(deco[nfiles][row])))
          if (read_float_column(infptr, "DEC_OFFSET", nrow, &ncol0,
                                &(deco[nfiles][row])))
            printerror(status);
        if (read_float_column(infptr, "Parallactic Angle", nrow, &ncol0,
                              &(pa[nfiles][row])))
          printerror(status);
        read_logical_column(infptr, "Scanning", nrow, &ncol0,
                             &(scanning[nfiles][row]));
        if (read_logical_column(infptr, "On Beam", nrow, &ncol0,
                             &(on_beam[nfiles][row])))
        {
          status = 0;
/* WORK HERE */
        }
        if (read_logical_column(infptr, "Off Beam", nrow, &ncol0,
                             &(off_beam[nfiles][row])))
        {
          status = 0;
/* WORK HERE */
        }
        if (read_float_column(infptr, "CHOP_OFFSET", nrow, &ncol0,
                              &(chop[nfiles][row])))
          printerror(status);
        if (read_double_column(infptr, "UT", nrow, &ncol0, &(ut[row])))
          printerror(status);
        if (read_double_column(infptr, "Detector Time", nrow, &ncol0,
                               &(det_time[row]))) {
          status = 0;
	  det_time[0] = -1.0e30;     /* flag */
        }
        row += nrow;
        hdunum++;
      }
      status = 0;
      if (fits_read_key(infptr, TFLOAT, "CDELT1", &sampdur, headline, &status))
          printerror(status);
      sampdur /= 1000.0;

      /* Go to pixels HDU; SHOULD SEARCH FOR HDU TYPE IN FUTURE */
      hdunum = 4;
      if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status));
        printerror(status);
      if (hdutype != BINARY_TBL)
        printabort("Data HDU of wrong type");
      if (fits_get_num_rows(infptr, &nrow, &status))
        printerror(status);
      offset[nfiles*MAX_O_TERMS] = (float *) malloc(ndet*sizeof(float));
      if (read_float_column(infptr, "Pixel Offsets", nrow, &ncol0,
                                 offset[nfiles*MAX_O_TERMS]))
        printerror(status);
      if ((nrow*ncol0) != ndet)
        printabort("Unexpected number of detectors");
      for (term = 1; term < MAX_O_TERMS; term++) {
        offset[nfiles*MAX_O_TERMS+term] =
                             (float *) malloc(ndet * sizeof(float));
        for (det = 1; det <= ndet; det++) {
          offset[nfiles*MAX_O_TERMS+term][det-1] = 0.0;
        }
      }

      if (fits_close_file(infptr, &status))
        printerror(status);

      badsamp[nfiles] = (char *) malloc(nsamp[nfiles] * sizeof(char));
      for (samp = 1; samp <= nsamp[nfiles]; samp++) {
        badsamp[nfiles][samp-1] = 0;
      }
      offset_cut(nsamp[nfiles], azo[nfiles], azerr[nfiles], elo[nfiles],
                 elerr[nfiles], el[nfiles], sampdur, badsamp[nfiles],
                 azo0[nfiles], -1.0*zao0[nfiles], ao_max, ao_min, eo_max,
                 eo_min, rao[nfiles], deco[nfiles], ro_max, ro_min, do_max,
                 do_min, year);
      sscanf(line, "%*s %n", &i);
      first = 'y';
      while (sscanf(&(line[i]), "%d %n", &samp, &j) == 1) {
        i += j;
        if (first == 'y') {
          firstsamp = samp;
          first = 'n';
        }
        else {
          for (j = firstsamp; j <= samp; j++) {
            if ((j >= 1) && (j <= nsamp[nfiles])) {
              badsamp[nfiles][j-1] = 1;
            }
          }
          first = 'y';
        }
      }
      if (azo_chop[nfiles] < 0.0)
        azo_chop[nfiles] *= -1.0;
      if (azo_chop[nfiles] > 0.1) {
        fprintf(stderr,
                "  NODDED OBSERVING MODE DETECTED -- APPLYING OFFSETS\n");
        for (samp = 1; samp <= nsamp[nfiles]; samp++) {
          if (on_beam[nfiles][samp-1]) {
            azo[nfiles][samp-1] -= azo_chop[nfiles];
          }
          else {     /* CHANGE THIS LATER TO if (off_beam[][])... */
            azo[nfiles][samp-1] += azo_chop[nfiles];
          }
        }
      }

      tauzen[nfiles] = tauzen0;
      skytran[nfiles] = (float *) malloc(nsamp[nfiles] * sizeof(float));
      calc_tran(skytran[nfiles], nsamp[nfiles], badsamp[nfiles], el[nfiles],
                tauzen[nfiles]);

      check_saturation(rawdata[nfiles], baddet, ndet, nsamp[nfiles],
                       badsamp[nfiles]);
      check_timing(ut, det_time, nsamp[nfiles], badsamp[nfiles], sampdur);
      free (ut);

      drop_list[nfiles] = drop;
      true_fazo[nfiles] = true_fazo0;
      true_fzao[nfiles] = true_fzao0;
      nfiles++;
     }
    }
    fclose(infile);
    if (newset != 'y')
      ngainsets++;

    /******** PRELIMINARY PROCESSING OF RAW DATA **********/

    fprintf(stderr, "\nPreliminary processing...\n");
    coadd_factor = 0.0;
    for (file = 0; file < nfiles; file++) {
      fprintf(stderr, "File %d\n", file+1);
      if (chop_flag == 'y') {
        calc_chop(chop[file], nsamp[file], chop_tol, &chop_hi, &chop_lo,
                  &chop_eff);
        fprintf(stderr,
                "Measured chop high, low (eff.):  %.1f, %.1f  (%.f%%)\n",
                chop_hi, chop_lo, chop_eff*100.0);
        fprintf(stderr, "Measured throw = %.1f arcsec\n",
                (chop_hi-chop_lo)*CHOP_SCALE);
        if (chop_throw < 0.0) {
          chop_throw = (chop_hi-chop_lo)*CHOP_SCALE;
        }
        else {
          fprintf(stderr, "FORCING throw = %.1f arcsec\n", chop_throw);
        }
        old_nsamp = nsamp[file];
        chop_demod(chop[file], &(nsamp[file]), chop_hi, chop_lo, chop_tol,
                   badsamp[file], &(rawdata[file]), ndet, &(el[file]),
                   &(ra[file]), &(dec[file]), &(azerr[file]), &(elerr[file]),
                   &(azo[file]), &(elo[file]), &(rao[file]), &(deco[file]),
                   &(pa[file]), &(scanning[file]));
        coadd_factor += old_nsamp/nsamp[file]*chop_eff;
        calc_chop_dx(nfiles, nsamp, pa, rotator, altaz_flag, chop_throw,
                     nasmyth_flag, &chop_dx, &chop_dy, chop_angle);
      }
      else {
        coadd_factor += 1.0;
      }
      for (term = 0; term < MAX_O_TERMS; term++) {
        basfunc[file*MAX_O_TERMS+term] =
                             (float *) malloc(nsamp[file] * sizeof(float));
      }
      calc_basfunc(&basfunc[file*MAX_O_TERMS], (o_order+d_order+1),
                   nsamp[file], badsamp[file], scanning[file],
                   scanning_flag);

      x0[file] = (float *) malloc(nsamp[file]*sizeof(float));
      y0[file] = (float *) malloc(nsamp[file]*sizeof(float));
      if (ignore_flag == 'n') {
        calc_xy_vs_time(x0[file], y0[file], nsamp[file], el[file], ra[file],
                        dec[file], azo[file], elo[file], azerr[file],
                        elerr[file], rao[file], deco[file], pa[file],
                        fixed_flag, true_fazo[file], true_fzao[file],
                        fazo[file], fzao[file], altaz_flag, rao0[file],
                        deco0[file], azo0[file], -1.0*zao0[file]);
      }
      else {
        for (samp = 1; samp <= nsamp[file]; samp++) {
          x0[file][samp-1] = 0.0;
          y0[file][samp-1] = 0.0;
        }
      }
      x1[file] = (float *) malloc(ndet*sizeof(float));
      y1[file] = (float *) malloc(ndet*sizeof(float));
      calc_rot_pixel_xy(x1[file], y1[file], ndet, nsamp[file], pa[file],
                    rotator[file], el[file], altaz_flag, nasmyth_flag);

      sign_correct(rawdata[file], offset[file*MAX_O_TERMS], ndet, nsamp[file]);

      f_wt[file] = (float *) malloc(nsamp[file] * sizeof(float));
      for (term = 0; term < B_TERMS; term++) {
        background[file*B_TERMS+term] =
                               (float *) malloc(nsamp[file] * sizeof(float));
        for (samp = 1; samp <= nsamp[file]; samp++) {
          background[file*B_TERMS+term][samp-1] = 0.0;
          f_wt[file][samp-1] = 1.0;     /* a little wasteful */
        }
      }
      for (term = 0; term < ARRAY_ROWS; term++) {
        rowavg[file*ARRAY_ROWS+term] =
                               (float *) malloc(nsamp[file] * sizeof(float));
        for (samp = 1; samp <= nsamp[file]; samp++) {
          rowavg[file*ARRAY_ROWS+term][samp-1] = 0.0;
        }
      }
      for (term = 0; term < MAX_O_TERMS; term++) {
        drift[file*MAX_O_TERMS+term] =
                               (float *) malloc(ARRAY_ROWS * sizeof(float));
        for (i = 0; i < ARRAY_ROWS; i++) {
          drift[file*MAX_O_TERMS+term][i] = 0.0;
        }
      }

      /* Get reasonable starting points for background and offsets: */
      pixel_mean_over_file(rawdata[file], gain[gset[file]],
                           offset[file*MAX_O_TERMS], background[file*B_TERMS],
                           ndet, nsamp[file], badsamp[file], x0[file],
                           y0[file], x1[file], y1[file], avoid_r2, mask,
                           chop_flag, chop_dx, chop_dy);
      mean_over_array(rawdata[file], gain[gset[file]], offset[file*MAX_O_TERMS],
                      background[file*B_TERMS], ndet, nsamp[file],
                      badsamp[file], baddet, x0[file], y0[file], x1[file],
                      y1[file], avoid_r2, mask, chop_flag, chop_dx, chop_dy);
      pixel_mean_over_file(rawdata[file], gain[gset[file]],
                           offset[file*MAX_O_TERMS], background[file*B_TERMS],
                           ndet, nsamp[file], badsamp[file], x0[file],
                           y0[file], x1[file], y1[file], avoid_r2, mask,
                           chop_flag, chop_dx, chop_dy);
      mean_over_array(rawdata[file], gain[gset[file]], offset[file*MAX_O_TERMS],
                      background[file*B_TERMS], ndet, nsamp[file],
                      badsamp[file], baddet, x0[file], y0[file], x1[file],
                      y1[file], avoid_r2, mask, chop_flag, chop_dx, chop_dy);
      if (filter_freq > 0.0) {
        fprintf(stderr, "Fourier filtering data above %.2f Hz.\n", filter_freq);
        pre_filter(rawdata[file], gain[gset[file]], offset[file*MAX_O_TERMS],
                   background[file*B_TERMS], ndet, nsamp[file], badsamp[file],
                   baddet, x0[file], y0[file], x1[file], y1[file], avoid_r2,
                   mask, filter_freq, sampdur);
      }
    }
    coadd_factor /= nfiles;
fprintf(stderr, "Coadd factor = %f\n", coadd_factor);
    image = (float *) malloc(IM_X*IM_Y*sizeof(float));
    weight = (float *) malloc(IM_X*IM_Y*sizeof(float));
    im_rms = (float *) malloc(IM_X*IM_Y*sizeof(float));
    free (mask);

    /************* ITERATE FOR OUTPUT IMAGE ***************/

    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        image[(y-1)*IM_X + (x-1)] = 0.0;
        weight[(y-1)*IM_X + (x-1)] = 0.0;
      }
    }
    gradient_flag = 'n';
    quadratic_flag = 'n';
    deep_flag = 'n';
    pwt_flag = 'n';
    mean = 0.0;
    for (iter = 1; iter <= num_iter; iter++) {
      if (iter >= (num_iter - map_iter + 1)) {
        fprintf(stderr, "src ");
        calc_map(image, weight, skytran, rawdata, gain, gset, p_wt, baddet,
                 badsamp, offset, basfunc, o_order, background, drift, d_order,
                 x0, y0, x1, y1, nfiles, nsamp, ndet, scanning, scanning_flag,
                 deep_flag, rowavg, &mean, rz_flag, zero_radius, chop_flag,
                 chop_dx, chop_dy);
        if (iter < (num_iter - faithful_iter + 1)) {
          fprintf(stderr, "neg ");
          remove_neg(image, neg_cutoff);
        }
      }
      if (iter == pwt_iter) {
        pwt_flag = 'y';
        fprintf(stderr, "pwt ");
      }
      if (iter > pwt_iter)
        pwt_flag = 'n';
      calc_rms(image, weight, im_rms, skytran, rawdata, gain, gset, p_wt,
               baddet, badsamp, offset, basfunc, o_order, background, f_wt,
               drift, d_order, x0, y0, x1, y1, nfiles, nsamp, ndet, scanning,
               scanning_flag, deep_flag, rowavg, mean, 'n', pwt_flag, 'n');
      fprintf(stderr, "Iteration #%2d:\n", iter);
      if (iter >= (num_iter - deep_iter + 1))
        deep_flag = 'y';
      if (iter >= (num_iter - grad_iter + 1))
        gradient_flag = 'y';
      if (iter >= (num_iter - quad_iter + 1))
        quadratic_flag = 'y';
      fprintf(stderr, "  off ");
      if (deep_flag == 'y') fprintf(stderr, "deep ");
      calc_offsets(image, weight, skytran, rawdata, gain, gset, baddet, badsamp,
                   offset, basfunc, o_order, background, drift, d_order, x0,
                   y0, x1, y1, nfiles, nsamp, ndet, scanning, scanning_flag,
                   deep_flag, rowavg);
      fprintf(stderr, "drft ");
      calc_drift(image, weight, skytran, rawdata, gain, gset, baddet, badsamp,
                 offset, basfunc, o_order, background, drift, d_order, x0, y0,
                 x1, y1, nfiles, nsamp, ndet, scanning, scanning_flag, rowavg);
      fprintf(stderr, "gain ");
      calc_gains(image, weight, skytran, rawdata, gain, gset, ngainsets, baddet,
                 badsamp, offset, basfunc, o_order, background, drift, d_order,
                 x0, y0, x1, y1, nfiles, nsamp, ndet, scanning, scanning_flag,
                 deep_flag, rowavg);
      norm_gains(gain, ngainsets, baddet, ndet);
      fprintf(stderr, "bgnd ");
      if (gradient_flag == 'y') fprintf(stderr, "grad ");
      if (quadratic_flag == 'y') fprintf(stderr, "quad ");
      calc_background(image, weight, skytran, rawdata, gain, gset, p_wt, baddet,
                      badsamp, offset, basfunc, o_order, background, drift,
                      d_order, x0, y0, x1, y1, nfiles, nsamp, ndet, scanning,
                      scanning_flag, gradient_flag, quadratic_flag, deep_flag,
                      rowavg);
      zero_background(background, nfiles, nsamp, badsamp, azo, azerr, el, elo,
                      elerr, rao, deco);
    }
    if (map_iter >= 0) {
      fprintf(stderr, "src ");
      calc_map(image, weight, skytran, rawdata, gain, gset, p_wt, baddet,
               badsamp, offset, basfunc, o_order, background, drift, d_order,
               x0, y0, x1, y1, nfiles, nsamp, ndet, scanning, scanning_flag,
               deep_flag, rowavg, &mean, rz_flag, zero_radius, chop_flag,
               chop_dx, chop_dy);
      rms = calc_rms(image, weight, im_rms, skytran, rawdata, gain, gset, p_wt,
                     baddet, badsamp, offset, basfunc, o_order, background,
                     f_wt, drift, d_order, x0, y0, x1, y1, nfiles, nsamp, ndet,
                     scanning, scanning_flag, deep_flag, rowavg, mean, 'y',
                     'y', 'n');
    }
    deg_count = 0;
    if (map_iter >= 0) {
      deglitch_map(image, im_rms, weight, skytran, rawdata, gain, gset, p_wt,
                   baddet, badsamp, offset, basfunc, o_order, background, drift,
                   d_order, x0, y0, x1, y1, nfiles, nsamp, ndet, scanning,
                   scanning_flag, deep_flag, rowavg, &mean, rz_flag,
                   zero_radius, chop_flag, chop_dx, chop_dy, deglitch_sigma,
                   &deg_count, drop_list);
    }
/* Following rms calculation doesn't use deglitching information and uses dropped scans*/
    rms = calc_rms(image, weight, im_rms, skytran, rawdata, gain, gset, p_wt,
                   baddet, badsamp, offset, basfunc, o_order, background, f_wt,
                   drift, d_order, x0, y0, x1, y1, nfiles, nsamp, ndet,
                   scanning, scanning_flag, deep_flag, rowavg, mean, 'y', 'y',
                   'y');

    good_det = 0.0;
    for (det = 1; det <= ndet; det++) {
      if ((baddet[det] & BAD_MASK) == 0)
        good_det += 1.0;
    }
    good_samp = 0.0;
    for (file = 0; file < nfiles; file++) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (scanning_flag || scanning[file][samp-1])
          if (badsamp[file][samp-1] == 0)
            good_samp += 1.0;
      }
    }
    fit_param = 0.0;
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (weight[(y-1)*IM_X + (x-1)] > 0.1)
          fit_param += 1.0;
      }
    }
    fit_param -= 1.0;     /* constraint of zero map sum */
    fit_param += good_det*nfiles;     /* gains */
    fit_param -= 1.0*nfiles;     /* constraint of unity gain average */
    fit_param += good_det*(o_order + 1.0)*nfiles;     /* offsets */
    if (deep_flag == 'y') {
      fit_param += ARRAY_ROWS*good_samp;     /* drifts */
    }
    else {
      fit_param += ARRAY_ROWS*d_order*nfiles;     /* drifts */
    }
    fit_param += 3.0*good_samp;     /* background (assumes gradient fits) */
    fit_param -= 3.0*nfiles;     /* constraint of zero background average */
    fprintf(stderr, "  %.0f measurements, %.0f fit parameters\n",
            (good_det*good_samp - deg_count), fit_param);
    fprintf(stderr, "  reduced chi2 = %f\n",
            rms*(good_det*good_samp)
            /(good_det*good_samp - deg_count - fit_param));

    /************* WRITE SOLUTION INFORMATION ***************/

    system("rm -f solve.txt");
    if ((outfile = fopen("solve.txt", "w")) == NULL)
      printabort("Couldn't write to solve.txt");
    fprintf(outfile, "!OVERALL RMS = %8.4f, reduced chi2 = %8.4f\n", rms,
            rms*(good_det*good_samp)/(good_det*good_samp - fit_param));

    /* pixel gains, status, and rms */
    worst_pixel = 0;
    worst_pixel_rms = 0.0;
    fprintf(outfile, "!PIXEL GAINS\n");
    fprintf(outfile, "!row column mean_gain bad_flag rms gains...\n");
    for (det = 1; det <= ndet; det++) {
      mean = 0.0;
      for (set = 0; set < ngainsets; set++)
        mean += gain[set][det];
      mean /= (float) ngainsets;
      fprintf(outfile, "%2d %2d  %7.4f %3d", (det-1)/ARRAY_COLUMNS + 1,
              (det-1)%ARRAY_COLUMNS + 1, mean, baddet[det]);
      mean = 1.0/sqrt(p_wt[det]);
      fprintf(outfile, "  %8.4f", mean);
      for (set = 0; set < ngainsets; set++)
        fprintf(outfile, "    %7.4f", gain[set][det]);
      fprintf(outfile, "\n");
      if ((mean > worst_pixel_rms) && ((baddet[det] & BAD_MASK) == 0)) {
        worst_pixel = det;
        worst_pixel_rms = mean;
      }
    }
    fprintf(stderr, "Worst pixel, rms = (%2d, %2d), %f\n",
            (worst_pixel-1)/ARRAY_COLUMNS+1,
            (worst_pixel-1)%ARRAY_COLUMNS+1, worst_pixel_rms);
    fprintf(outfile, "!WORST PIXEL, RMS = (%2d, %2d), %f\n",
            (worst_pixel-1)/ARRAY_COLUMNS+1,
            (worst_pixel-1)%ARRAY_COLUMNS+1, worst_pixel_rms);

    /* pixel offsets */
    fprintf(outfile, "!OFFSET ORDER = %d\n", o_order);
    fprintf(outfile, "!PIXEL OFFSET TERMS\n");
    fprintf(outfile, "!row column terms\n");
    for (det = 1; det <= ndet; det++) {
      fprintf(outfile, "%2d %2d  ", (det-1)/ARRAY_COLUMNS + 1,
              (det-1)%ARRAY_COLUMNS + 1);
      for (file = 0; file < nfiles; file++) {
        for (term = 0; term <= o_order; term++) {
          fprintf(outfile, " %9.4f", offset[file*MAX_O_TERMS+term][det-1]);
        }
      }
      fprintf(outfile, "\n");
    }

    /* row drifts */
    fprintf(outfile, "!DRIFT ORDER = %d\n", d_order);
    fprintf(outfile, "!ROW DRIFT TERMS\n");
    fprintf(outfile, "!file row terms\n");
    for (file = 0; file < nfiles; file++) {
      for (i = 0; i < ARRAY_ROWS; i++) {
        fprintf(outfile, "%2d %2d  ", file+1, i+1);
        for (term = 0; term < d_order; term++) {
          fprintf(outfile, " %9.4f", drift[file*MAX_O_TERMS+term][i]);
        }
        fprintf(outfile, "\n");
      }
    }

    /* frame rms */
    worst_frame = 0;
    worst_frame_rms = 0.0;
    fprintf(outfile, "!BACKGROUND AND FRAME RMS\n");
    fprintf(outfile, "!file frame background rms\n");
    for (file = 0; file < nfiles; file++) {
      sum3 = 0.0;
      sum4 = 0.0;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        fprintf(outfile, "%2d %5d ", file+1, samp);
        for (term = 0; term < B_TERMS; term++) {
          fprintf(outfile, " %9.4f", background[file*B_TERMS+term][samp-1]);
        }
        if (scanning_flag || scanning[file][samp-1]) {
          sum0 = 0.0;
          sum1 = 0.0;
          for (det = 1; det <= ndet; det++) {
            if ((baddet[det] & BAD_MASK) == 0) {
              get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x,
                     &y, badsamp[file][samp-1]);
              sum2 = rawdata[file][(samp-1)*ndet + (det-1)];
              o = get_offset(&offset[file*MAX_O_TERMS],
                             &basfunc[file*MAX_O_TERMS], o_order,
                             nsamp[file], samp, det);
              d = get_drift(&drift[file*MAX_O_TERMS],
                            &basfunc[file*MAX_O_TERMS], o_order, d_order,
                            nsamp[file], samp, det, deep_flag,
                            &rowavg[file*ARRAY_ROWS]);
              b = get_background(&background[file*B_TERMS], samp, det);
              sum2 -= o;
              sum2 -= d;
              sum2 /= gain[gset[file]][det];
              if (x > 0)
                sum2 -= skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
              sum2 -= b;
              sum1 += sum2*sum2;
              sum0 += 1.0;
              if (badsamp[file][samp-1] == 0) {
                sum4 += sum2*sum2;
                sum3 += 1.0;
              }
            }
          }
          mean = sqrt(sum1/sum0);
          if (badsamp[file][samp-1] != 0)
            mean *= -1.0;
          fprintf(outfile, " %8.4f", mean);
          if (mean > worst_frame_rms) {
            worst_frame = samp;
            worst_frame_file = file;
            worst_frame_rms = mean;
          }
        }
        fprintf(outfile, "\n");
      }
      mean = sqrt(sum4/sum3);
      fprintf(outfile, "!FILE %2d rms = %8.4f\n", file+1, mean);
    }
    fprintf(stderr, "Worst file, frame, rms = %d, %d, %f\n", worst_frame_file+1,
            worst_frame, worst_frame_rms);
    fprintf(outfile, "!WORST FILE, FRAME, RMS = %d, %d, %f\n",
            worst_frame_file+1, worst_frame, worst_frame_rms);
    max_weight = 0.0;
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (weight[(y-1)*IM_X + (x-1)] > max_weight)
          max_weight = weight[(y-1)*IM_X + (x-1)];
      }
    }
    fprintf(stderr, "Maximum image weight = %.0f\n", max_weight);
    fprintf(outfile, "!MAXIMUM IMAGE WEIGHT = %.0f\n", max_weight);
    fclose(outfile);

    /*************** MODIFY RAW DATA, PART 1 -- RESIDUAL ****************/

    for (file = 0; file < nfiles; file++) {
      for (det = 1; det <= ndet; det++) {
        if ((baddet[det] & BAD_MASK) == 0) {
          for (samp = 1; samp <= nsamp[file]; samp++) {
            if (badsamp[file][samp-1] == 0) {
              o = get_offset(&offset[file*MAX_O_TERMS],
                             &basfunc[file*MAX_O_TERMS], o_order, nsamp[file],
                             samp, det);
              d = get_drift(&drift[file*MAX_O_TERMS],
                            &basfunc[file*MAX_O_TERMS], o_order, d_order,
                            nsamp[file], samp, det, deep_flag,
                            &rowavg[file*ARRAY_ROWS]);
              b = get_background(&background[file*B_TERMS], samp, det);
              rawdata[file][(samp-1)*ndet + (det-1)] -= (o+d);
              rawdata[file][(samp-1)*ndet + (det-1)] /= gain[gset[file]][det];
              rawdata[file][(samp-1)*ndet + (det-1)] -= b;
              get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x,
                     &y, badsamp[file][samp-1]);
              if (x > 0)
                rawdata[file][(samp-1)*ndet + (det-1)] -=
                             skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
            }
            else{
              rawdata[file][(samp-1)*ndet + (det-1)] = 0.0;
            }
          }
        }
        else {
          for (samp = 1; samp <= nsamp[file]; samp++) {
            rawdata[file][(samp-1)*ndet + (det-1)] = 0.0;
          }
        }
      }
    }

    /*************** COMPUTE COVARIANCE MATRICES ****************/

    if (cov_flag == 'y') {
      /* detector X detector */
      fprintf(stderr, "Calculating detector X detector covariance matrix\n");
      covar = (float *) malloc(ndet*ndet*sizeof(float));
      for (det1 = 1; det1 <= ndet; det1++) {
        for (det2 = 1; det2 <= ndet; det2++) {
          covar[(det1 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2 - 1)] = 0.0;
        }
      }
      for (det1 = 1; det1 <= ndet; det1++) {
        if ((baddet[det1] & BAD_MASK) == 0) {
          for (det2 = det1; det2 <= ndet; det2++) {
            if ((baddet[det2] & BAD_MASK) == 0) {
              for (file = 0; file < nfiles; file++) {
                for (samp = 1; samp <= nsamp[file]; samp++) {
                  if (badsamp[file][samp-1] == 0) {
                    covar[(det1 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2 - 1)]
                         += rawdata[file][(samp-1)*ndet + (det1-1)]
                           *rawdata[file][(samp-1)*ndet + (det2-1)];
                    if (det1 != det2) {
                      covar[(det2 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1 - 1)]
                           += rawdata[file][(samp-1)*ndet + (det1-1)]
                             *rawdata[file][(samp-1)*ndet + (det2-1)];
                    }
                  }
                }
              }
            }
          }
        }
      }
      /* Normalize off-diagonal elements */
      for (det1 = 1; det1 < ndet; det1++) {
        if ((baddet[det1] & BAD_MASK) == 0) {
          for (det2 = (det1+1); det2 <= ndet; det2++) {
            if ((baddet[det2] & BAD_MASK) == 0) {
              covar[(det1 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2 - 1)]
                  /= sqrt(covar[(det1-1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1-1)]);
              covar[(det1 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2 - 1)]
                  /= sqrt(covar[(det2-1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2-1)]);
              covar[(det2 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1 - 1)]
                  /= sqrt(covar[(det1-1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1-1)]);
              covar[(det2 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1 - 1)]
                  /= sqrt(covar[(det2-1)*ARRAY_ROWS*ARRAY_COLUMNS + (det2-1)]);
            }
          }
        }
      }
      /* Normalize on-diagonal elements */
      for (det1 = 1; det1 <= ndet; det1++) {
        if ((baddet[det1] & BAD_MASK) == 0) {
          covar[(det1 - 1)*ARRAY_ROWS*ARRAY_COLUMNS + (det1 - 1)] = 1.0;
        }
      }
      system("rm -f detector_detector.fits");
      if (fits_create_file(&outfptr, "detector_detector.fits", &status))
        printerror(status);
      bitpix = FLOAT_IMG;
      naxis = 3;
      naxes[0] = ARRAY_COLUMNS;
      naxes[1] = ARRAY_ROWS;
      naxes[2] = ndet;
      nelements = naxes[0]*naxes[1]*naxes[2];
      fpixel[0] = 1;
      fpixel[1] = 1;
      fpixel[2] = 1;
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, covar,
                         &status))
        printerror(status);
      naxis = 2;
      naxes[0] = ndet;
      naxes[1] = ndet;
      nelements = naxes[0]*naxes[1];
      fpixel[0] = 1;
      fpixel[1] = 1;
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, covar,
                         &status))
        printerror(status);
      if (fits_close_file(outfptr, &status))
        printerror(status);
      free (covar);
    }

    /******** MODIFY RAW DATA, PART 2 -- RESTORE SIGNAL ********/

    if (residual_flag != 'y') {
      for (file = 0; file < nfiles; file++) {
        for (det = 1; det <= ndet; det++) {
          if ((baddet[det] & BAD_MASK) == 0) {
            for (samp = 1; samp <= nsamp[file]; samp++) {
              if (badsamp[file][samp-1] == 0) {
                get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x,
                       &y, badsamp[file][samp-1]);
                if (x > 0)
                  rawdata[file][(samp-1)*ndet + (det-1)] +=
                             skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
              }
            }
          }
        }
      }
    }

    /************** CLIP, SCALE, AND LEVEL IMAGE **************/

    if ((chop_flag == 'y') && (chop_dual_flag == 'n')) {
      chop_combine(image, weight, im_rms, chop_dx, chop_dy);
    }

    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (weight[(y-1)*IM_X + (x-1)] < (weight_cut*max_weight)) {
          image[(y-1)*IM_X + (x-1)] = 0.0;
          weight[(y-1)*IM_X + (x-1)] = 0.0;
          im_rms[(y-1)*IM_X + (x-1)] = 0.0;
        }
        else {
          image[(y-1)*IM_X + (x-1)] *= mul_const;
          image[(y-1)*IM_X + (x-1)] += zero;
          im_rms[(y-1)*IM_X + (x-1)] *= mul_const;
          weight[(y-1)*IM_X + (x-1)] *= coadd_factor;
        }
      }
    }

    /************* WRITE OUTPUT IMAGE ***************/

    system("rm -f solve.fits");
    if (fits_create_file(&outfptr, "solve.fits", &status))
      printerror(status);
    bitpix = FLOAT_IMG;

    if (output_flag == 'd') {
      /* pixels vs. time (dump) image */
      for (file = 0; file < nfiles; file++) {
        if (avg_flag == 'y')
          block_average(rawdata[file], ndet, &nsamp[file]);
        naxis = 2;
        naxes[0] = ndet;
        naxes[1] = nsamp[file];
        if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
          printerror(status);
        fpixel[0] = 1;
        fpixel[1] = 1;
        nelements = ndet*nsamp[file];
        if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, rawdata[file],
                           &status))
          printerror(status);
      }
    }

    if (output_flag == 's') {
      /* sky plane image */
      naxis = 2;
      naxes[0] = IM_X;
      naxes[1] = IM_Y;
      nelements = IM_X*IM_Y;
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      fpixel[0] = 1;
      fpixel[1] = 1;
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, image, &status))
        printerror(status);
      val = equinox;
      if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &val, "", &status))
        printerror(status);
      if (fits_write_key(outfptr, TSTRING, "CTYPE1", "RA---SIN", "", &status))
        printerror(status);
      val = raep*15.0;
      if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &val, "", &status))
        printerror(status);
      val = SCALE/MAGNIFICATION/-3600.0;
      if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
        printerror(status);
      val = IM_X0;
      if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
        printerror(status);
      val = 0.0;
      if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
        printerror(status);
      if (fits_write_key(outfptr, TSTRING, "CTYPE2", "DEC--SIN", "", &status))
        printerror(status);
      val = decep;
      if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &val, "", &status))
        printerror(status);
      val = SCALE/MAGNIFICATION/3600.0;
      if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
        printerror(status);
      val = IM_Y0;
      if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
        printerror(status);
      val = 0.0;
      if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
        printerror(status);

      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Weight",
                         "Identification of the HDU", &status))
        printerror(status);
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, weight, &status))
        printerror(status);

      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Noise",
                         "Identification of the HDU", &status))
        printerror(status);
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, im_rms, &status))
        printerror(status);

      for (x = 1; x <= IM_X; x++) {
        for (y = 1; y <= IM_Y; y++) {
          if (im_rms[(y-1)*IM_X + (x-1)] > 0.0) {
            im_rms[(y-1)*IM_X + (x-1)] = image[(y-1)*IM_X + (x-1)]
                                       / im_rms[(y-1)*IM_X + (x-1)];
          }
        }
      }
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
        printerror(status);
      if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Signal-to-Noise",
                         "Identification of the HDU", &status))
        printerror(status);
      if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, im_rms, &status))
        printerror(status);
    }

    if (output_flag == 'm') {
      /* movie */
      for (file = 0; file < nfiles; file++) {
        if (avg_flag == 'y')
          block_average(rawdata[file], ndet, &nsamp[file]);
        naxis = 3;
        naxes[0] = ARRAY_COLUMNS;
        naxes[1] = ARRAY_ROWS;
        naxes[2] = nsamp[file];
        if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
          printerror(status);
        fpixel[0] = 1;
        fpixel[1] = 1;
        fpixel[2] = 1;
        nelements = ndet*nsamp[file];
        if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, rawdata[file],
                           &status))
          printerror(status);
      }
    }

    /* Assume at last HDU, then copy headers to new appended HDU's */

    if ((infile = fopen(argv[1], "r")) == NULL)
      printabort("Could not open list of FITS files (second pass)");
    file = 0;
    while (fgets(line, MAX_LEN, infile) != NULL) {
      sscanf(line, "%s", filename);     /* Has terminating 0 */
      if (filename[0] != '$') {
        if (drop_list[file] == 0) {
          if (fits_open_file(&infptr, filename, READONLY, &status))
            printerror(status);
          if (fits_copy_hdu(infptr, outfptr, 0, &status))
            printerror(status);
          if (fits_close_file(infptr, &status))
            printerror(status);
        }
        file++;
      }
    }
    fclose(infile);

    /* Return to HDU 0 to write solution options */
    hdunum = 1;
    if (fits_movabs_hdu(outfptr, hdunum, &hdutype, &status));
      printerror(status);
    if (fits_write_key(outfptr, TSTRING, "RED_VER", version,
                       "Reduction code version", &status))
      printerror(status);
    time(&the_time);
    sprintf(line, "  reduced %s", ctime(&the_time));
    line[strlen(line)-1] = '\0';    /* overwrite '\a' */
    if (fits_write_comment(outfptr, line, &status))
      printerror(status);
    sprintf(line, "skymap");
    if (output_flag == 'd') {
      if (residual_flag == 'y') {
        sprintf(line, "residual dump");
      }
      else {
        sprintf(line, "dump");
      }
    }
    if (output_flag == 'm') {
      if (residual_flag == 'y') {
        sprintf(line, "residual movie");
      }
      else {
        sprintf(line, "movie");
      }
    }
    val = DISTORT;
    if (fits_write_key(outfptr, TFLOAT, "DISTORT", &val,
                       "distortion = y scale / x scale (-di)", &status))
      printerror(status);
    if (fits_write_key(outfptr, TSTRING, "OUT_FORM", line,
                       "Output image format (-du, -mo, -er)", &status))
      printerror(status);
    ival = num_iter;
    if (fits_write_key(outfptr, TINT, "NUM_ITER", &ival,
                       "Total iterations (-n)", &status))
      printerror(status);
    ival = map_iter;
    if (fits_write_key(outfptr, TINT, "MAP_ITER", &ival,
                       "Iterations with nonzero map (-i)", &status))
      printerror(status);
    ival = faithful_iter;
    if (fits_write_key(outfptr, TINT, "FTH_ITER", &ival,
                       "Iterations without negative clipping (-ft)", &status))
      printerror(status);
    val = neg_cutoff;
    if (fits_write_key(outfptr, TFLOAT, "NEG_CUT", &val,
                       "Neg. clip. lev. for unfaithful iter. (-fl)", &status))
      printerror(status);
    ival = grad_iter;
    if (fits_write_key(outfptr, TINT, "GRA_ITER", &ival,
                       "Iterations with background gradients (-g)", &status))
      printerror(status);
    ival = quad_iter;
    if (fits_write_key(outfptr, TINT, "QUA_ITER", &ival,
                       "Iterations with background quadratic (-q)", &status))
      printerror(status);
    ival = deep_iter;
    if (fits_write_key(outfptr, TINT, "DEP_ITER", &ival,
                       "Iterations with deep mode (-de)", &status))
      printerror(status);
    ival = pwt_iter;
    if (fits_write_key(outfptr, TINT, "PWT_ITER", &ival,
                       "Iteration with pixel weight recalc. (-p)", &status))
      printerror(status);
    ival = o_order;
    if (fits_write_key(outfptr, TINT, "OFF_ORD", &ival,
                       "Pixel offset order (-o)", &status))
      printerror(status);
    ival = d_order;
    if (fits_write_key(outfptr, TINT, "DRIF_ORD", &ival,
                       "Row drift additional order (-o)", &status))
      printerror(status);
    if (filter_freq > 0.0) {
      val = filter_freq;
      if (fits_write_key(outfptr, TFLOAT, "FILT_FRQ", &val,
                         "Minimum filter frequency (-fi)", &status))
        printerror(status);
    }
    val = deglitch_sigma;
    if (fits_write_key(outfptr, TFLOAT, "DEGL_SIG", &val,
                       "Number of sigma for deglitching (-dg)", &status))
      printerror(status);
    sprintf(line, "%c", altaz_flag);
    if (fits_write_key(outfptr, TSTRING, "ALTAZ", line,
                       "Alt-az coordinates for sky map? (-al)", &status))
      printerror(status);
    if (fixed_flag == 'y') {
      for (i = 0; i < nfiles; i++) {
        sprintf(line, "FAZO_N%03d", i+1);
        val = true_fazo0;
        if (fits_write_key(outfptr, TFLOAT, line, &val,
                           "New value of FAZO (-fa)", &status))
          printerror(status);
        sprintf(line, "FZAO_N%03d", i+1);
        val = true_fzao0;
        if (fits_write_key(outfptr, TFLOAT, line, &val,
                           "New value of FZAO (-fa)", &status))
          printerror(status);
      }
    }
    val = avoid_r2;
    if (val > 0.0)
      val = sqrt(val);
    if (fits_write_key(outfptr, TFLOAT, "AVOID", &val,
                       "Avoidance radius (pixels) for 0'th iter. (-ra)",
                       &status))
      printerror(status);
    sprintf(line, "%c", mask_flag);
    if (fits_write_key(outfptr, TSTRING, "MASK", line,
                       "Use mask for 0'th iteration? (-ms)", &status))
      printerror(status);
    if (mask_flag == 'y') {
      if (fits_write_key(outfptr, TSTRING, "MASKFILE", maskfilename,
                         "Mask file name (-ms), contents below:", &status))
        printerror(status);
      if ((infile = fopen(maskfilename, "r")) == NULL) {
        sprintf(line, "Couldn't open mask file %s", maskfilename);
        printabort(line);
      }
      while (fgets(line, MAX_LEN, infile) != NULL) {
        line[strlen(line)-1] = '\0';    /* overwrite '\a' */
        if (fits_write_comment(outfptr, line, &status))
          printerror(status);
      }
      fclose(infile);
    }
    if (rz_flag == 'y') {
      val = zero_radius;
      if (fits_write_key(outfptr, TFLOAT, "ZERO_RAD", &val,
                         "Radius for map zeroing (-rz)", &status))
        printerror(status);
    }
    if (r0_flag == 'y') {
      val = ROTATOR_0;
      if (fits_write_key(outfptr, TFLOAT, "ROTZERO", &val,
                         "Rotator angle (deg.) for altaz alignment (-r0)",
                         &status))
        printerror(status);
    }
    if (ro_flag == 'y') {
      val = rotator_override;
      if (fits_write_key(outfptr, TFLOAT, "ROT_OVER", &val,
                         "Forced rotator angle (deg.) (-ro)", &status))
        printerror(status);
    }
    for (i = 0; i < nfiles; i++) {
      val = tauzen[i];    /* WORK HERE -- record all values */
      sprintf(line, "TAU_Z%03d", i+1);
      if (fits_write_key(outfptr, TFLOAT, line, &val,
                         "Zenith opacity used in solution (-t,-cf)", &status))
        printerror(status);
    }
    val = ao_min;
    if (fits_write_key(outfptr, TFLOAT, "AZO_MIN", &val,
                       "Minimum azimuth scan offset (-ac)", &status))
      printerror(status);
    val = ao_max;
    if (fits_write_key(outfptr, TFLOAT, "AZO_MAX", &val,
                       "Maximum azimuth scan offset (-ac)", &status))
      printerror(status);
    val = eo_min;
    if (fits_write_key(outfptr, TFLOAT, "ELO_MIN", &val,
                       "Minimum elevation scan offset (-ec)", &status))
      printerror(status);
    val = eo_max;
    if (fits_write_key(outfptr, TFLOAT, "ELO_MAX", &val,
                       "Maximum elevation scan offset (-ec)", &status))
      printerror(status);
    val = ro_min;
    if (fits_write_key(outfptr, TFLOAT, "RAO_MIN", &val,
                       "Minimum RA scan offset (-rc)", &status))
      printerror(status);
    val = ro_max;
    if (fits_write_key(outfptr, TFLOAT, "RAO_MAX", &val,
                       "Maximum RA scan offset (-rc)", &status))
      printerror(status);
    val = do_min;
    if (fits_write_key(outfptr, TFLOAT, "DECO_MIN", &val,
                       "Minimum dec. scan offset (-dc)", &status))
      printerror(status);
    val = do_max;
    if (fits_write_key(outfptr, TFLOAT, "DECO_MAX", &val,
                       "Maximum dec. scan offset (-dc)", &status))
      printerror(status);
    if (chop_flag == 'y') {
      val = chop_throw;
      if (fits_write_key(outfptr, TFLOAT, "CHOP_THR", &val,
                         "Chop throw in arcsec (-ch,-cl)", &status))
        printerror(status);
      val = chop_tol;
      if (fits_write_key(outfptr, TFLOAT, "CHOP_TOL", &val,
                         "Chop tolerance in arcsec (-ct)", &status))
        printerror(status);
    }
    if (output_flag == 's') {
      val = weight_cut;
      if (fits_write_key(outfptr, TFLOAT, "MIN_WGT", &val,
                         "Minimum fractional weight for sky image (-w)",
                         &status))
        printerror(status);
      val = mul_const;
      if (fits_write_key(outfptr, TFLOAT, "MUL_CNST", &val,
                         "Mult. constant applied to sky image (-mu, -cf)",
                         &status))
        printerror(status);
      val = zero;
      if (fits_write_key(outfptr, TFLOAT, "ADD_CNST", &val,
                         "Constant added to sky image (-z)", &status))
        printerror(status);
    }
    if (avg_flag == 'y') {
      ival = BLOCK_SIZE;
      if (fits_write_key(outfptr, TINT, "BLOCK", &ival,
                         "Block size for dump and movie (-b)", &status))
        printerror(status);
    }

    if (fits_close_file(outfptr, &status))
      printerror(status);

}

/*******************/

int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                      float *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    float  nulval = FLOATNULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TFLOAT) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(float)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                       double *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    double  nulval = DOUBLENULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TDOUBLE) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(double)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                        char *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    char  nulval = 0;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TLOGICAL) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(char)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,
               float *chop_lo, float *chop_eff)

{
    float  himean, lomean, hisig, losig;
    float  himax, himin, lomax, lomin;
    float  mean, sum0, hi0, lo0;
    long  samp;
    int  i;

    /* Should I exclude bad samples? */
    mean = 0.0;
    sum0 = 0.0;
    for (samp = 1; samp <= nsamp; samp++) {
      mean += chop[samp-1];
      sum0 += 1.0;
    }
    mean /= sum0;
    himax = 1.0e30;
    himin = mean;
    lomin = -1.0e30;
    lomax = mean;
    for (i = 1; i <= 10; i++) {
      himean = 0.0;
      lomean = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
       if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          himean += chop[samp-1];
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          lomean += chop[samp-1];
          lo0 += 1.0;
        }
      }
      himean /= hi0;
      lomean /= lo0;
      hisig = 0.0;
      losig = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          hisig += (chop[samp-1] - himean)*(chop[samp-1] - himean);
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          losig += (chop[samp-1] - lomean)*(chop[samp-1] - lomean);
          lo0 += 1.0;
        }
      }
      hisig = sqrt(hisig/hi0);
      losig = sqrt(losig/lo0);
      if (hisig < (chop_tol / 2.0))
        hisig = chop_tol / 2.0;
      if (losig < (chop_tol / 2.0))
        losig = chop_tol / 2.0;
      himax = himean + hisig;
      himin = himean - hisig;
      lomax = lomean + losig;
      lomin = lomean - losig;
    }
    *chop_hi = himean;
    *chop_lo = lomean;
    *chop_eff = (hi0 + lo0)/sum0;
    /* assumes that sigmas eventually become smaller than chop_tol / 2.0 */
}

/*******************/

void sign_correct(float *rawdata, float offset[], long ndet, long nsamp)

{
    long  det, samp;

    for (det = 1; det <= ndet; det++) {
      for (samp = 1; samp <= nsamp; samp++) {
        rawdata[(samp-1)*ndet + (det-1)] *= -1000.0;
        /* Convert to mV and invert */
      }
      offset[det-1] *= -1000.0;
      /* Convert to mV and invert */
    }
}

/*******************/

void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,
                      char *badsamp)

{
    long  det, samp;
    char  ok_flag[ARRAY_ROWS*ARRAY_COLUMNS];
    float  raw;

    for (det = 1; det <= ndet; det++)
      ok_flag[det-1] = 'y';
    for (det = 1; det <= ndet; det++) {
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          raw = rawdata[(samp-1)*ndet + (det-1)];
          if ((raw > MAX_V) || (raw < MIN_V))
            ok_flag[det-1] = 'n';
        }
      }
    }
    for (det = 1; det <= ndet; det++) {
      if ((ok_flag[det-1] != 'y') && ((baddet[det] & BAD_MASK) == 0)) {
        fprintf(stderr,
                "  WARNING:  Detector (%2d, %2d) near saturation\n",
                (det-1)/ARRAY_COLUMNS+1, (det-1)%ARRAY_COLUMNS+1);
        system("date >> sharcsolve_errlog.txt");
        system(
        "echo \"WARNING:  Detector near saturation\" >> sharcsolve_errlog.txt");
      }
    }
}

/*******************/

void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,
                  float sampdur)

{
    long  samp, minsamp, maxsamp;
    double  dut, mindut, maxdut;

    mindut = 1.0e30;
    maxdut = -1.0e30;
    for (samp = 1; samp < nsamp; samp++) {
      if ((badsamp[samp-1] == 0) && (badsamp[samp] == 0)) {
        dut = ut[samp] - ut[samp-1];
        if (dut < mindut) {
          mindut = dut;
          minsamp = samp+1;
        }
        if (dut > maxdut) {
          maxdut = dut;
          maxsamp = samp+1;
        }
      }
    }
    if (mindut < 0.0) {
      fprintf(stderr, "  WARNING:  Time not monotonic for sample %ld\n",
              minsamp);
      system("date >> sharcsolve_errlog.txt");
      system("echo \"WARNING:  Time not monotonic\" >> sharcsolve_errlog.txt");
    }
    maxdut *= 3600.0;
    if (maxdut > (MAX_TIME_JUMP*sampdur)) {
      fprintf(stderr, "  WARNING:  Time jump (%.3lf sec) for sample %ld\n",
              maxdut, maxsamp);
      system("date >> sharcsolve_errlog.txt");
      system("echo \"WARNING:  Time jump\" >> sharcsolve_errlog.txt");
    }
    mindut = 1.0e30;
    maxdut = -1.0e30;
    if (det_time[0] > -1.0e29) {
      for (samp = 1; samp <= nsamp; samp++) {
        if ((badsamp[samp-1] == 0)) {
          dut = det_time[samp-1] - ut[samp-1]*3600.0;     /* seconds */
          if (dut < mindut) {
            mindut = dut;
            minsamp = samp+1;
          }
          if (dut > maxdut) {
            maxdut = dut;
            maxsamp = samp+1;
          }
        }
      }
      dut = (maxdut + mindut)/2.0;
      dut = dut - 86400.0*((int) (dut/86400.0 + 0.5));
      fprintf(stderr, "  Post April 2003 timing check:  %.3f %.3f %.3f %.3f\n",
              mindut, maxdut, dut, maxdut - mindut);
      if (fabs(dut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization offset:  %.3f sec\n", dut);
        system("date >> sharcsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization offset\" >> sharcsolve_errlog.txt");
      }
      if ((maxdut - mindut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization glitch:  %.3f sec\n", maxdut-mindut);
        system("date >> sharcsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization glitch\" >> sharcsolve_errlog.txt");
      }
    }
}

/*******************/

void offset_cut(long nsamp, float *azo, float *azerr, float *elo,
                float *elerr, float *el, float sampdur, char *badsamp,
                float  azo0, float elo0, float ao_max, float ao_min,
                float eo_max, float eo_min, float *rao, float *deco,
                float ro_max, float ro_min, float do_max, float do_min,
                int year)

{
    long  samp, csamp;
    float  daz, del;
    float  azmin = 1.0e30;
    float  azmax =-1.0e30;
    float  elmin = 1.0e30;
    float  elmax =-1.0e30;
    float  ramin = 1.0e30;
    float  ramax =-1.0e30;
    float  decmin = 1.0e30;
    float  decmax =-1.0e30;

    /* Note not using azerr and elerr information */
    csamp = 0;
    for (samp = 1; samp <= nsamp; samp++) {
      if (year < 2003) {
        azo[samp-1] -= azo0;    /* Reconstruct sweep offset */
        elo[samp-1] -= elo0;
      }
      daz = azo[samp-1];
      del = elo[samp-1];
      if ((daz >= ao_max) || (daz <= ao_min) ||
          (del >= eo_max) || (del <= eo_min)) {
        badsamp[samp-1] = 1;
        csamp++;
      }
      else {
        if ((rao[samp-1] >= ro_max) || (rao[samp-1] <= ro_min) ||
            (deco[samp-1] >= do_max) || (deco[samp-1] <= do_min)) {
          badsamp[samp-1] = 1;
          csamp++;
        }
      }
      if (daz < azmin)
        azmin = daz;
      if (daz > azmax)
        azmax = daz;
      if (del < elmin)
        elmin = del;
      if (del > elmax)
        elmax = del;
      if (rao[samp-1] < ramin)
        ramin = rao[samp-1];
      if (rao[samp-1] > ramax)
        ramax = rao[samp-1];
      if (deco[samp-1] < decmin)
        decmin = deco[samp-1];
      if (deco[samp-1] > decmax)
        decmax = deco[samp-1];
    }
    fprintf(stderr,
        "  Maximum sweep offsets:  %.1f %.1f %.1f %.1f %.1f %.1f %.1f %.1f\n",
        azmin, azmax, elmin, elmax, ramin, ramax, decmin, decmax);
    if (csamp > 0)
      fprintf(stderr, "  Cut %ld samples based on offsets\n", csamp);
}

/*******************/

void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,
                float chop_tol, char *badsamp, float **rawdata, long ndet,
                float **el, float **ra, float **dec, float **azerr,
                float **elerr, float **azo, float **elo, float **rao,
                float **deco, float **pa, char **scanning)

{
    long  *choptag;
    float  *hiweight, *loweight;
    float  himax, himin, lomax, lomin;
    long  hicount, locount;
    int  old_chop_state, new_chop_state;
    long  samp, nchop, samp0, samp1;
    float  *c_rawdata, *c_el, *c_ra, *c_dec, *c_azerr, *c_elerr;
    float  *c_azo, *c_elo, *c_rao, *c_deco, *c_pa;
    char  *c_scanning;
    long  det;

    /* DOESN'T CORRECTLY USE SCANNING FLAG */
    /* SHOULD HANDLE PA DIFFERENTLY */

    choptag = (long *) malloc(*nsamp*sizeof(long));
    hiweight = (float *) malloc(*nsamp*sizeof(float));
    loweight = (float *) malloc(*nsamp*sizeof(float));

    himax = chop_hi + chop_tol/2.0;
    himin = chop_hi - chop_tol/2.0;
    lomax = chop_lo + chop_tol/2.0;
    lomin = chop_lo - chop_tol/2.0;
    hicount = 0;
    locount = 0;
    old_chop_state = 0;
    nchop = 0;
    samp0 = 1;
    for (samp = 1; samp <= *nsamp; samp++) {
      new_chop_state = 0;
      hiweight[samp-1] = 0.0;
      loweight[samp-1] = 0.0;
      if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax))
        new_chop_state = 1;
      if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax))
        new_chop_state = -1;
      if (new_chop_state != 0) {
        if ((new_chop_state*old_chop_state) == -1) {
          if ((locount > (hicount/2)) && (hicount > (locount/2))) {
            for (samp1 = samp0; samp1 < samp; samp1++) {
              hiweight[samp1-1] /= (float)hicount;
              loweight[samp1-1] /= (float)locount;
            }
            hicount = 0;
            locount = 0;
            nchop++;
            samp0 = samp;
          }
          else {
            for (samp1 = samp0; samp1 < samp; samp1++) {
              if (new_chop_state == 1)
                hiweight[samp1-1] = 0.0;
              if (new_chop_state == -1)
                loweight[samp1-1] = 0.0;
            }
            if (new_chop_state == 1)
              hicount = 0;
            if (new_chop_state == -1)
              locount = 0;
          }
        }
        if (badsamp[samp-1] == 0) {
          if (new_chop_state == 1) {
            hiweight[samp-1] = 1.0;
            hicount++;
          }
          if (new_chop_state == -1) {
            loweight[samp-1] = 1.0;
            locount++;
          }
        }
        old_chop_state = new_chop_state;
      }
      choptag[samp-1] = nchop;
    }
    if ((locount > (hicount/2)) && (hicount > (locount/2))) {
      for (samp1 = samp0; samp1 < samp; samp1++) {
        hiweight[samp1-1] /= (float)hicount;
        loweight[samp1-1] /= (float)locount;
      }
      hicount = 0;
      locount = 0;
      nchop++;
      samp0 = samp;
    }

    c_rawdata = (float *) malloc(nchop*ndet*sizeof(float));
    c_el = (float *) malloc(nchop*sizeof(float));
    c_ra = (float *) malloc(nchop*sizeof(float));
    c_dec = (float *) malloc(nchop*sizeof(float));
    c_azerr = (float *) malloc(nchop*sizeof(float));
    c_elerr = (float *) malloc(nchop*sizeof(float));
    c_azo = (float *) malloc(nchop*sizeof(float));
    c_elo = (float *) malloc(nchop*sizeof(float));
    c_rao = (float *) malloc(nchop*sizeof(float));
    c_deco = (float *) malloc(nchop*sizeof(float));
    c_pa = (float *) malloc(nchop*sizeof(float));
    c_scanning = (char *) malloc(nchop*sizeof(char));

    for (samp = 1; samp <= nchop; samp++) {
      for (det = 1; det <= ndet; det++) {
        c_rawdata[(samp-1)*ndet + (det-1)] = 0.0;
      }
      c_el[samp-1] = 0.0;
      c_ra[samp-1] = 0.0;
      c_dec[samp-1] = 0.0;
      c_azerr[samp-1] = 0.0;
      c_elerr[samp-1] = 0.0;
      c_azo[samp-1] = 0.0;
      c_elo[samp-1] = 0.0;
      c_rao[samp-1] = 0.0;
      c_deco[samp-1] = 0.0;
      c_pa[samp-1] = 0.0;
      c_scanning[samp-1] = 1;
    }

    for (samp = 1; samp <= *nsamp; samp++) {
      if (choptag[samp-1] < nchop) {
        for (det = 1; det <= ndet; det++) {
          c_rawdata[choptag[samp-1]*ndet + (det-1)] +=
                                     (*rawdata)[(samp-1)*ndet + (det-1)] *
                                     (hiweight[samp-1]-loweight[samp-1]);
        }
        c_el[choptag[samp-1]] += (*el)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_ra[choptag[samp-1]] += (*ra)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_dec[choptag[samp-1]] += (*dec)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azerr[choptag[samp-1]] += (*azerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elerr[choptag[samp-1]] += (*elerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azo[choptag[samp-1]] += (*azo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elo[choptag[samp-1]] += (*elo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_rao[choptag[samp-1]] += (*rao)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_deco[choptag[samp-1]] += (*deco)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_pa[choptag[samp-1]] += (*pa)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
      }
    }

    free(*rawdata);
    *rawdata = c_rawdata;
    free(*el);
    *el = c_el;
    free(*ra);
    *ra = c_ra;
    free(*dec);
    *dec = c_dec;
    free(*azerr);
    *azerr = c_azerr;
    free(*elerr);
    *elerr = c_elerr;
    free(*azo);
    *azo = c_azo;
    free(*elo);
    *elo = c_elo;
    free(*rao);
    *rao = c_rao;
    free(*deco);
    *deco = c_deco;
    free(*pa);
    *pa = c_pa;
    free(*scanning);
    *scanning = c_scanning;

    *nsamp = nchop;
    free(choptag);
    free(hiweight);
    free(loweight);
}

/*******************/

void pixel_ntile_over_file(float *rawdata, float offset[], long ndet,
                           long nsamp, char *badsamp)

/* NOT CURRENT -- MAY CRASH */

{
    long  det, samp;
    float  datum;
    float  list[MAX_LIST], ntile;
    int  listsize, k;

    for (det = 1; det <= ndet; det++) {
      listsize = 0;
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          datum = rawdata[(samp-1)*ndet + (det-1)];
          listsize++;
          list[listsize] = datum;
        }
      }
      k = (listsize*NTILE)/100;
      ntile = element_select(k, listsize, list);
      offset[det-1] = ntile;
    }
}

/*******************/

void pixel_mean_over_file(float *rawdata, float gain[], float offset[],
                          float *background, long ndet, long nsamp,
                          char *badsamp, float *x0, float *y0, float *x1,
                          float *y1, float avoid_r2, char *mask,
                          char chop_flag, int chop_dx, int chop_dy)

/* uses only uniform background term */

/* Should I incorporate scanning and scanning_flag? */

{
    long  det, samp;
    int  x, y;
    float  datum;
    double  sum0, sum1, mean;
    float  r2, s2;
    int  dx, dy;

    if (chop_flag == 'y') {
      dx = chop_dx;
      dy = chop_dy;
    }
    else {
      dx = 0;
      dy = 0;
    }

    /* subtract individual offsets */
    for (det = 1; det <= ndet; det++) {
      sum0 = 0.0;
      sum1 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          get_xy(det, samp, x0, y0, x1, y1, &x, &y, badsamp[samp-1]);
          r2 = (x-IM_X0-dx)*(x-IM_X0-dx) + (y-IM_Y0-dy)*(y-IM_Y0-dy);
          s2 = (x-IM_X0+dx)*(x-IM_X0+dx) + (y-IM_Y0+dy)*(y-IM_Y0+dy);
          if ((r2 > avoid_r2) && (s2 > avoid_r2) && (x > 0)) {
            if (mask[(y-1)*IM_X + (x-1)] == 0) {
              datum = rawdata[(samp-1)*ndet + (det-1)]
                      -background[samp-1]*gain[det];
              sum0 += 1.0;
              sum1 += datum;
            }
          }
        }
      }
      if (sum0 < 0.1) {
        offset[det-1] = 0.0;
      }
      else {
        mean = sum1/sum0;
        offset[det-1] = mean;
      }
    }
}

/*******************/

void ntile_over_array(float *rawdata, float *gain, float offset[],
                      float *background, long ndet, long nsamp, int baddet[])

/* NOT CURRENT -- MAY CRASH */

{
    long  det, samp;
    float  datum;
    float  list[MAX_LIST], ntile;
    int  listsize, k;

    for (samp = 1; samp <= nsamp; samp++) {
      listsize = 0;
      for (det = 1; det <= ndet; det++) {
        if ((baddet[det] & BAD_MASK) == 0) {
          listsize++;
          list[listsize] = (rawdata[(samp-1)*ndet + (det-1)]
                            -offset[det-1]) / gain[det];
        }
      }
      k = (listsize*NTILE)/100;
      ntile = element_select(k, listsize, list);
      background[samp-1] = ntile;
    }
}

/*******************/

void mean_over_array(float *rawdata, float gain[], float offset[],
                     float *background, long ndet, long nsamp, char *badsamp,
                     int baddet[], float *x0, float *y0, float *x1, float *y1,
                     float avoid_r2, char *mask, char chop_flag, int chop_dx,
                     int chop_dy)

{
    long  det, samp, first_good_samp;
    char  first_flag;
    int  x, y;
    double  sum0, sum1, sum2, sum3, mean;
    float  r2, s2;
    int  dx, dy;

     if (chop_flag == 'y') {
      dx = chop_dx;
      dy = chop_dy;
    }
    else {
      dx = 0;
      dy = 0;
    }

    sum2 = 0.0;
    sum3 = 0.0;
    first_good_samp = 1;
    first_flag = 'n';
    for (samp = 1; samp <= nsamp; samp++) {
      sum0 = 0.0;
      sum1 = 0.0;
      for (det = 1; det <= ndet; det++) {
        if ((baddet[det] & BAD_MASK) == 0) {
          get_xy(det, samp, x0, y0, x1, y1, &x, &y, badsamp[samp-1]);
          r2 = (x-IM_X0-dx)*(x-IM_X0-dx) + (y-IM_Y0-dy)*(y-IM_Y0-dy);
          s2 = (x-IM_X0+dx)*(x-IM_X0+dx) + (y-IM_Y0+dy)*(y-IM_Y0+dy);
          if ((r2 > avoid_r2) && (s2 > avoid_r2) && (x > 0)) {
            if (mask[(y-1)*IM_X + (x-1)] == 0) {
              sum0 += 1.0;
              sum1 += (rawdata[(samp-1)*ndet + (det-1)] -
                       offset[det-1]) / gain[det];
            }
          }
        }
      }
      if (sum0 < 0.1) {
        if (first_flag == 'n') {
          first_good_samp++;
          mean = 0.0;
        }
        else {
          mean = background[samp-2];
        }
      }
      else {
        first_flag = 'y';
        mean = sum1/sum0;
      }
      background[samp-1] = mean;
      if (badsamp[samp-1] == 0) {
        sum2 += 1.0;
        sum3 += mean;
      }
    }
    for (samp = 1; samp < first_good_samp; samp++) {
      background[samp-1] = background[first_good_samp-1];
      sum3 += background[first_good_samp-1];
      /* Don't need to add to sum2 */
    }
    mean = sum3/sum2;
    for (samp = 1; samp <= nsamp; samp++) {
      background[samp-1] -= mean;
    }
}

/*******************/

void pre_filter(float *rawdata, float gain[], float offset[],
                float *background, long ndet, long nsamp, char *badsamp,
                int baddet[], float *x0, float *y0, float *x1, float *y1,
                float avoid_r2, char *mask, float filter_freq, float sampdur)

{
    unsigned long  n, n1;
    long  det, samp;
    float  *data;

    /* Note no bad frame checking */
    n = log(nsamp - 0.5)/log(2.0);
    n = pow(2.0, n + 1.0) + 0.5;     /* power of 2 greater than or equal to */
    n1 = n*filter_freq*2.0*sampdur;
    data = vector(1, n);
    for (det = 1; det <= ndet; det++) {
      for (samp = 1; samp <= nsamp; samp++)
        data[samp] = rawdata[(samp-1)*ndet + (det-1)] - offset[det-1]
                     - background[samp-1]*gain[det];
      for (samp = nsamp + 1; samp <= n; samp++)
        data[samp] = 0.0;     /* zero pad */
      realft(data, n, 1);     /* F.T. */
      for (samp = n1; samp <= n; samp++)
        data[samp] = 0.0;     /* filter high frequencies */
      realft(data, n, -1);     /* inverse F.T. */
      for (samp = 1; samp <= n; samp++)
        data[samp] *= (2.0/n);     /* normalize */
      for (samp = 1; samp <= nsamp; samp++)
        rawdata[(samp-1)*ndet + (det-1)] = data[samp] + offset[det-1]
                     + background[samp-1]*gain[det];
    }
    free_vector(data, 1, n);
}

/*******************/

void calc_xy_vs_time(float *x0, float *y0, long nsamp, float *el, float *ra,
                     float *dec, float *azo, float *elo, float *azerr,
                     float *elerr, float *rao, float *deco, float *pa,
                     char fixed_flag, float true_fazo, float true_fzao,
                     float fazo, float fzao, char altaz_flag, float rao0,
                     float deco0, float azo0, float elo0)

{
    long  samp;
    float  daz, del, dx, dy;
    float  spa, cpa;

    /* Note that ra[] and dec[] are ignored */

    /* using samp index 0 to nsamp-1 */
    for (samp = 0; samp < nsamp; samp++) {
      daz = azo[samp] + azo0 + azerr[samp]*cos(el[samp]*PI/180.0);
      del = elo[samp] + elo0 + elerr[samp];
      if (fixed_flag == 'y') {
        daz += (fazo - true_fazo);
        del -= (fzao - true_fzao);
      }
      if (altaz_flag == 'y') {
        dx = daz;
        dy = del;
      }
      else {
        cpa = cos(pa[samp]*PI/180.0);
        spa = sin(pa[samp]*PI/180.0);
        dx = daz*cpa - del*spa - rao[samp] - rao0;
        dy = daz*spa + del*cpa + deco[samp] + deco0;
      }
      x0[samp] = dx/SCALE*MAGNIFICATION;
      y0[samp] = dy/SCALE*MAGNIFICATION;
    }
}

/*******************/

void calc_pixel_xy(long det, float *x, float *y)

{
    int  xx, yy;

    xx = (det - 1) % ARRAY_COLUMNS;
    yy = (det - 1) / ARRAY_COLUMNS;

    *x = (A_CEN_X - xx)/DISTORT;
    *y = (yy - A_CEN_Y)*DISTORT;
}

/*******************/

void calc_rot_pixel_xy(float *x1, float *y1, long ndet, long nsamp, float *pa,
                       float rotator, float *el, char altaz_flag,
                       char nasmyth_flag)

{
    long  samp, det;
    double  sum0;
    float  mean, cpa, spa, dx, dy, dxx, dyy;

    /* compute mean parallactic angle */
    sum0 = 0.0;
    cpa = 0.0;
    spa = 0.0;
    for (samp = 0; samp < nsamp; samp++) {
      sum0 += 1.0;
      cpa += cos(pa[samp]*PI/180.0);
      spa += sin(pa[samp]*PI/180.0);
    }
    cpa /= sum0;
    spa /= sum0;
    mean = atan2(spa, cpa)*180.0/PI;
    fprintf(stderr, "nsamp = %d, ", nsamp);
    fprintf(stderr, "Mean PA = %f\n", mean);
    if (altaz_flag == 'y') {
      fprintf(stderr, "  FORCING to PA = 0 for altaz mode\n");
      mean = 0.0;
    }
    mean -= rotator;
    if (nasmyth_flag == 'y') {
      mean += (el[0] + el[nsamp-2])/2.0;
    }
    fprintf(stderr, "  Mean array rotation = %f\n", mean);
    cpa = cos(mean*PI/180.0);
    spa = sin(mean*PI/180.0);

    for (det = 1; det <= ndet; det++) {
      calc_pixel_xy(det, &dx, &dy);
      dxx = dx*cpa - dy*spa;
      dyy = dx*spa + dy*cpa;
      x1[det-1] = dxx*MAGNIFICATION;
      y1[det-1] = dyy*MAGNIFICATION;
    }
}

/*******************/

void calc_chop_dx(int nfiles, long nsamp[], float *pa[], float rotator[],
                  char altaz_flag, float chop_throw, char nasmyth_flag,
                  int *chop_dx, int *chop_dy, float chop_angle)

{
    int  file;
    long  samp;
    double  sum0;
    float  mean, angle, cpa, spa;

    cpa = 0.0;
    spa = 0.0;
    sum0 = 0.0;
    for (file = 0; file < nfiles; file++) {
      for (samp = 0; samp < nsamp[file]; samp++) {
        sum0 += 1.0;
        angle = 0.0;
        if (altaz_flag != 'y') {
          angle += pa[file][samp];
        }
        cpa += cos(angle*PI/180.0);
        spa += sin(angle*PI/180.0);
      }
    }
    cpa /= sum0;
    spa /= sum0;
    mean = atan2(spa, cpa) + chop_angle*PI/180.0;
    cpa = cos(mean);     /* normalize */
    spa = sin(mean);
    *chop_dx = floor(cpa*chop_throw*MAGNIFICATION/SCALE/2.0+0.5);
    *chop_dy = floor(spa*chop_throw*MAGNIFICATION/SCALE/2.0+0.5);
fprintf(stderr, "CHOP DX, DY = %d, %d\n", *chop_dx, *chop_dy);
}

/*******************/

void chop_combine(float *image, float *weight, float *im_rms, int dx, int dy)

{
    int  x, y;
    float  *comb_image, *comb_weight, *comb_rms;


    comb_image = (float *) malloc(IM_X*IM_Y*sizeof(float));
    comb_weight = (float *) malloc(IM_X*IM_Y*sizeof(float));
    comb_rms = (float *) malloc(IM_X*IM_Y*sizeof(float));

    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (   ((x - dx) >= 1) && ((x - dx) <= IM_X)
            && ((x + dx) >= 1) && ((x + dx) <= IM_X)
            && ((y - dy) >= 1) && ((y - dy) <= IM_Y)
            && ((y + dy) >= 1) && ((y + dy) <= IM_Y)) {
          if (   (weight[(y-dy-1)*IM_X + (x-dx-1)] > 0.1)
               &&(weight[(y+dy-1)*IM_X + (x+dx-1)] > 0.1)) {
            comb_image[(y-1)*IM_X + (x-1)] =
              (image[(y-dy-1)*IM_X + (x-dx-1)]
               - image[(y+dy-1)*IM_X + (x+dx-1)])/2.0;
            comb_weight[(y-1)*IM_X + (x-1)] =
              1.0/(1.0/weight[(y-dy-1)*IM_X + (x-dx-1)]
                   + 1.0/weight[(y+dy-1)*IM_X + (x+dx-1)]);
            comb_rms[(y-1)*IM_X + (x-1)] =
              sqrt( im_rms[(y-dy-1)*IM_X + (x-dx-1)]
	           *im_rms[(y-dy-1)*IM_X + (x-dx-1)]
                  + im_rms[(y+dy-1)*IM_X + (x+dx-1)]
                   *im_rms[(y+dy-1)*IM_X + (x+dx-1)])/2.0;
          }
          else {
            comb_image[(y-1)*IM_X + (x-1)] = 0.0;
            comb_weight[(y-1)*IM_X + (x-1)] = 0.0;
            comb_rms[(y-1)*IM_X + (x-1)] = 0.0;
          }
        }
        else {
          comb_image[(y-1)*IM_X + (x-1)] = 0.0;
          comb_weight[(y-1)*IM_X + (x-1)] = 0.0;
          comb_rms[(y-1)*IM_X + (x-1)] = 0.0;
        }
      }
    }
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
         image[(y-1)*IM_X + (x-1)] = comb_image[(y-1)*IM_X + (x-1)];
         weight[(y-1)*IM_X + (x-1)] = comb_weight[(y-1)*IM_X + (x-1)];
         im_rms[(y-1)*IM_X + (x-1)] = comb_rms[(y-1)*IM_X + (x-1)];
      }
    }

    free(comb_image);
    free(comb_weight);
    free(comb_rms);
}

/*******************/

void get_xy(long pix, long samp, float *x0, float *y0, float *x1, float *y1,
            int *x, int *y, char bad_flag)

{
    float  dx, dy;

    if ((pix < 1) || (pix > ARRAY_ROWS*ARRAY_COLUMNS))
      printabort("Pixel out of range");
    dx = x0[samp-1] + x1[pix-1];
    dy = y0[samp-1] + y1[pix-1];
    *x = dx + IM_X0 + 0.5;     /* assuming truncation */
    *y = dy + IM_Y0 + 0.5;
    if ((*x < 1) || (*x > IM_X) || (*y < 1) || (*y > IM_Y)) {
      if (bad_flag == 0) {
        if (map_too_small != 'y') {
          fprintf(stderr, "WARNING:  Map not big enough:  x,y = %d, %d\n",
                  *x, *y);
          system("date >> sharcsolve_errlog.txt");
          system(
            "echo \"WARNING:  Map not big enough\" >> sharcsolve_errlog.txt");
        }
        map_too_small = 'y';
      }
      *x = -9999;
      *y = -9999;
    }
}

/*******************/

void calc_basfunc(float *basfunc[], int order, long nsamp, char *badsamp,
                  char *scanning, char scanning_flag)

{
    long  samp;
    double  *dbasfunc[MAX_O_TERMS];
    double  x;
    double  au, al, bl, a, b;
    int  term;

    /* Use method in section 4.5 of Numerical Recipes to generate
       orthogonal set of polynomials */

    for (term = 0; term < order; term++) {
      dbasfunc[term] = (double *) malloc(nsamp*sizeof(double));
    }

    /* polynomial 0 */
    au = 0.0;
    al = 0.0;
    for (samp = 1; samp <= nsamp; samp++) {
      x = (samp - 1.0)/(nsamp - 1.0);
      dbasfunc[0][samp-1] = 1.0;
      basfunc[0][samp-1] = 1.0;
      if (badsamp[samp-1] == 0) {
        if (scanning_flag || scanning[samp-1]) {
          au += x;
          al += 1.0;
        }
      }
    }
    a = au/al;
    bl = al;

    /* polynomial 1 */
    if (order > 1) {
      au = 0.0;
      al = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        x = (samp - 1.0)/(nsamp - 1.0);
        dbasfunc[1][samp-1] = x - a;
        if (badsamp[samp-1] == 0) {
          if (scanning_flag || scanning[samp-1]) {
            au += x*dbasfunc[1][samp-1]*dbasfunc[1][samp-1];
            al += dbasfunc[1][samp-1]*dbasfunc[1][samp-1];
          }
        }
      }
      a = au/al;
      b = al/bl;
      bl = al;
    }

    /* polynomials 2+ */
    for (term = 2; term < order; term++) {
      au = 0.0;
      al = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        x = (samp - 1.0)/(nsamp - 1.0);
        dbasfunc[term][samp-1] = (x - a)*dbasfunc[term-1][samp-1]
                                -b*dbasfunc[term-2][samp-1];
        if (badsamp[samp-1] == 0) {
          if (scanning_flag || scanning[samp-1]) {
            au += x*dbasfunc[term][samp-1]*dbasfunc[term][samp-1];
            al += dbasfunc[term][samp-1]*dbasfunc[term][samp-1];
          }
        }
      }
      a = au/al;
      b = al/bl;
      bl = al;
    }
    /* NOTE:  These polynomials become very small for large index */

    /* make orthonormal, except 0th term */
    for (term = 1; term < order; term++) {
      al = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          if (scanning_flag || scanning[samp-1]) {
            al += dbasfunc[term][samp-1]*dbasfunc[term][samp-1];
          }
        }
      }
      al = sqrt(al);
      for (samp = 1; samp <= nsamp; samp++)
        basfunc[term][samp-1] = dbasfunc[term][samp-1]/al;
      free(dbasfunc[term]);
    }
    free(dbasfunc[0]);
}

/*******************/

void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],
               float tauzen)

{
    long  samp;

    for (samp = 1; samp <= nsamp; samp++) {
      skytran[samp-1] = exp(-1.0*tauzen/sin(el[samp-1]*PI/180.0));
      if (el[samp-1] < 0.1) {
        fprintf(stderr, "  WARNING:  Bad elevation in frame %d -- flagging\n",
                samp);
        badsamp[samp-1] = 1;
        skytran[samp-1] = 1.0e-6;
        system("date >> sharcsolve_errlog.txt");
        system("echo \"WARNING:  Bad elevation\" >> sharcsolve_errlog.txt");
      }
    }
}

/*******************/

float get_offset(float *offset[], float *basfunc[], int o_order, long nsamp,
                 long samp, long det)

{
    int  term;
    double  dbo;
    float  fo;

    dbo = 0.0;
    for (term = 0; term <= o_order; term++) {
      dbo += offset[term][det-1]*basfunc[term][samp-1];
    }
    fo = dbo;

    return(fo);
}

/*******************/

float get_drift(float *drift[], float *basfunc[], int o_order, int d_order,
                long nsamp, long samp, long det, char deep_flag,
                float *rowavg[])

{
    int  term;
    float  fd;
    int  row;

    row = (det-1)/ARRAY_COLUMNS;
    if (deep_flag == 'y') {
      fd = rowavg[row][samp-1];
    }
    else {
      fd = 0.0;
      for (term = 0; term < d_order; term++) {
        fd += drift[term][row]*basfunc[term+o_order+1][samp-1];
      }
    }
    return(fd);
}

/*******************/

float get_background(float *background[], long samp, long det)

{
    float  b, dx, dy;

    calc_pixel_xy(det, &dx, &dy);
    b = background[0][samp-1];
    b += dx*background[1][samp-1];
    b += dy*background[2][samp-1];
    b += dx*dx*background[3][samp-1];

    return (b);
}

/*******************/

void calc_map(float *image, float *weight, float *skytran[], float *rawdata[],
              float *gain[], int gset[], float p_wt[], int baddet[],
              char *badsamp[], float *offset[], float *basfunc[], int o_order,
              float *background[], float *drift[], int d_order, float *x0[],
              float *y0[], float *x1[], float *y1[], int nfiles, long nsamp[],
              long ndet, char *scanning[], char scanning_flag, char deep_flag,
              float *rowavg[], float *mean, char rz_flag, float zero_radius,
              char chop_flag, int chop_dx, int chop_dy)

{
    int  file;
    long  samp, det;
    int  x, y;
    float  o, d, b, im;
    double  sum0, sum1;
    float  avg_wt;
    float  r2, s2, zr2;
    int  dx, dy;

    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        image[(y-1)*IM_X + (x-1)] = 0.0;
        weight[(y-1)*IM_X + (x-1)] = 0.0;
      }
    }

    /* This is already sky space minimization, as opposed to voltage space */
    for (file = 0; file < nfiles; file++) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (scanning_flag || scanning[file][samp-1]) {
          if (badsamp[file][samp-1] == 0) {
            for (det = 1; det <= ndet; det++) {
              if ((det % ARRAY_COLUMNS) == 1) {
                d = get_drift(&drift[file*MAX_O_TERMS],
                              &basfunc[file*MAX_O_TERMS], o_order, d_order,
                              nsamp[file], samp, det, deep_flag,
                              &rowavg[file*ARRAY_ROWS]);
              }
              if ((baddet[det] & BAD_MASK) == 0) {
                o = get_offset(&offset[file*MAX_O_TERMS],
                               &basfunc[file*MAX_O_TERMS], o_order,
                               nsamp[file], samp, det);
                get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                       &x, &y, badsamp[file][samp-1]);
                b = get_background(&background[file*B_TERMS], samp, det);
                im = ((rawdata[file][(samp-1)*ndet + (det-1)] - o - d)
                      /gain[gset[file]][det] - b) / skytran[file][samp-1];
                if (x > 0) {
                  image[(y-1)*IM_X + (x-1)] += im*p_wt[det];
                  weight[(y-1)*IM_X + (x-1)] += 1.0*p_wt[det];
                }
              }
            }
          }
        }
      }
    }
    avg_wt = avg_p_wt(p_wt, baddet, ndet);

    if (chop_flag == 'y') {
      dx = chop_dx;
      dy = chop_dy;
    }
    else {
      dx = 0;
      dy = 0;
    }
    if (rz_flag == 'y') {
      zr2 = zero_radius*zero_radius;
      for (x = 1; x <= IM_X; x++) {
        for (y = 1; y <= IM_Y; y++) {
          if (weight[(y-1)*IM_X + (x-1)] > (0.1*avg_wt))
            image[(y-1)*IM_X + (x-1)] /= weight[(y-1)*IM_X + (x-1)];
          weight[(y-1)*IM_X + (x-1)] /= avg_wt;
          r2 = (x-IM_X0-dx)*(x-IM_X0-dx) + (y-IM_Y0-dy)*(y-IM_Y0-dy);
          s2 = (x-IM_X0+dx)*(x-IM_X0+dx) + (y-IM_Y0+dy)*(y-IM_Y0+dy);
          if ((r2 > zr2) && (s2 > zr2)) {
            image[(y-1)*IM_X + (x-1)] = 0.0;
            weight[(y-1)*IM_X + (x-1)] = 0.0;
          }
        }
      }
      *mean = 0.0;
    }
    else {
      /* calculate mean of map */
      sum0 = 0.0;
      sum1 = 0.0;
      for (x = 1; x <= IM_X; x++) {
        for (y = 1; y <= IM_Y; y++) {
          if (weight[(y-1)*IM_X + (x-1)] > (0.1*avg_wt)) {
            image[(y-1)*IM_X + (x-1)] /= weight[(y-1)*IM_X + (x-1)];
            sum1 += image[(y-1)*IM_X + (x-1)];
            sum0 += 1.0;
          }
        }
      }
      *mean = sum1/sum0;
      /* subtract mean from map */
      for (x = 1; x <= IM_X; x++) {
        for (y = 1; y <= IM_Y; y++) {
          if (weight[(y-1)*IM_X + (x-1)] > (0.1*avg_wt)) {
            image[(y-1)*IM_X + (x-1)] -= *mean;
            weight[(y-1)*IM_X + (x-1)] /= avg_wt;
          }
        }
      }
    }
}

/*******************/

void remove_neg(float *image, float neg_cutoff)

{
    int  x, y;

fprintf(stderr, "Clipping negatives...\n");
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (image[(y-1)*IM_X + (x-1)] < neg_cutoff) {
          image[(y-1)*IM_X + (x-1)] = neg_cutoff;
        }
      }
    }
}

/*******************/

float calc_rms(float *image, float *weight, float *im_rms, float *skytran[],
              float *rawdata[], float *gain[], int gset[], float p_wt[],
              int baddet[], char *badsamp[], float *offset[], float *basfunc[],
              int o_order, float *background[], float *f_wt[], float *drift[],
              int d_order, float *x0[], float *y0[], float *x1[], float *y1[],
              int nfiles, long nsamp[], long ndet, char *scanning[],
              char scanning_flag, char deep_flag, float *rowavg[], float mean,
              char calc_im_rms, char calc_p_rms, char calc_f_rms)

{
    int  file;
    long  samp, det;
    int  x, y;
    float  o, d, b, im;
    double  sum0, sum1, sum2;
    float  overall_rms;
    float  avg_wt;

/* NEED TO CHECK RMS CALCULATIONS */
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        im_rms[(y-1)*IM_X + (x-1)] = 0.0;
      }
    }
    sum0 = 0.0;
    sum1 = 0.0;
    for (file = 0; file < nfiles; file++) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (scanning_flag || scanning[file][samp-1]) {
          if (badsamp[file][samp-1] == 0) {
            for (det = 1; det <= ndet; det++) {
              if ((det % ARRAY_COLUMNS) == 1) {
                d = get_drift(&drift[file*MAX_O_TERMS],
                              &basfunc[file*MAX_O_TERMS], o_order, d_order,
                              nsamp[file], samp, det, deep_flag,
                              &rowavg[file*ARRAY_ROWS]);
              }
              if ((baddet[det] & BAD_MASK) == 0) {
                o = get_offset(&offset[file*MAX_O_TERMS],
                               &basfunc[file*MAX_O_TERMS], o_order,
                               nsamp[file], samp, det);
                b = get_background(&background[file*B_TERMS], samp, det);
                get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x,
                       &y, badsamp[file][samp-1]);
                if (x > 0) {
                  im = ((rawdata[file][(samp-1)*ndet + (det-1)] - o - d)
                        /gain[gset[file]][det] - b) / skytran[file][samp-1];
                  sum2 = im - (image[(y-1)*IM_X + (x-1)] + mean);
                  sum1 += sum2*sum2*skytran[file][samp-1]*skytran[file][samp-1]
                          *p_wt[det];
                  sum0 += 1.0*p_wt[det];
                  if (calc_im_rms == 'y')
                    im_rms[(y-1)*IM_X + (x-1)] += sum2*sum2*p_wt[det];
                }
              }
            }
          }
        }
      }
    }
    overall_rms = sqrt(sum1/sum0);
    fprintf(stderr, "RMS = %f\n", overall_rms);

    if (calc_im_rms == 'y') {
      avg_wt = avg_p_wt(p_wt, baddet, ndet);
      for (x = 1; x <= IM_X; x++) {
        for (y = 1; y <= IM_Y; y++) {
          if (weight[(y-1)*IM_X + (x-1)] > 1.0)
            im_rms[(y-1)*IM_X + (x-1)] =
                 sqrt(im_rms[(y-1)*IM_X + (x-1)] / weight[(y-1)*IM_X + (x-1)]
                      / (weight[(y-1)*IM_X + (x-1)] - 1.0) / avg_wt);
            /* standard deviation of the mean */
        }
      }
    }

    if (calc_p_rms == 'y') {
      for (det = 1; det <= ndet; det++) {
        sum0 = 0.0;
        sum1 = 0.0;
        for (file = 0; file < nfiles; file++) {
          for (samp = 1; samp <= nsamp[file]; samp++) {
            if (scanning_flag || scanning[file][samp-1]) {
              if (badsamp[file][samp-1] == 0) {
                o = get_offset(&offset[file*MAX_O_TERMS],
                               &basfunc[file*MAX_O_TERMS], o_order, nsamp[file],
                               samp, det);
                d = get_drift(&drift[file*MAX_O_TERMS],
                              &basfunc[file*MAX_O_TERMS], o_order, d_order,
                              nsamp[file], samp, det, deep_flag,
                              &rowavg[file*ARRAY_ROWS]);
                get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x,
                       &y, badsamp[file][samp-1]);
                if (x > 0) {
                  b = get_background(&background[file*B_TERMS], samp, det);
                  sum2 = rawdata[file][(samp-1)*ndet + (det-1)] - o - d;
                  sum2 /= gain[gset[file]][det];
                  sum2 -= skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
                  sum2 -= b;
                  sum1 += sum2*sum2;
                  sum0 += 1.0;
                }
              }
            }
          }
        }
        p_wt[det] = sum0/sum1;    /* weight = 1/rms^2 */
      }
    }

    return(overall_rms);
}

/*******************/

void deglitch_map(float *image, float *im_rms, float *weight, float *skytran[],
                  float *rawdata[], float *gain[], int gset[], float p_wt[],
                  int baddet[], char *badsamp[], float *offset[],
                  float *basfunc[], int o_order, float *background[],
                  float *drift[], int d_order, float *x0[], float *y0[],
                  float *x1[], float *y1[], int nfiles, long nsamp[], long ndet,
                  char *scanning[], char scanning_flag, char deep_flag,
                  float *rowavg[], float *mean, char rz_flag,
                  float zero_radius, char chop_flag, int chop_dx, int chop_dy,
                  float deg_thresh, long *deg_count, char drop_list[])

{
    int  file;
    long  samp, det;
    int  x, y;
    float  o, d, b, im, rms;
    float  *oldim, *oldwt;
    double  sum0, sum1;
    float  avg_wt;

/* NEED TO CHECK DEGLITCHING TEST */
    oldim = (float *) malloc(IM_X*IM_Y*sizeof(float));
    oldwt = (float *) malloc(IM_X*IM_Y*sizeof(float));
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        oldim[(y-1)*IM_X + (x-1)] = image[(y-1)*IM_X + (x-1)] + *mean;
        image[(y-1)*IM_X + (x-1)] = 0.0;
        oldwt[(y-1)*IM_X + (x-1)] = weight[(y-1)*IM_X + (x-1)];
        weight[(y-1)*IM_X + (x-1)] = 0.0;
      }
    }

    *deg_count = 0;
    for (file = 0; file < nfiles; file++) {
     if (drop_list[file] == 0) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (scanning_flag || scanning[file][samp-1]) {
          if (badsamp[file][samp-1] == 0) {
            for (det = 1; det <= ndet; det++) {
              if ((det % ARRAY_COLUMNS) == 1) {
                d = get_drift(&drift[file*MAX_O_TERMS],
                              &basfunc[file*MAX_O_TERMS], o_order, d_order,
                              nsamp[file], samp, det, deep_flag,
                              &rowavg[file*ARRAY_ROWS]);
              }
              if ((baddet[det] & BAD_MASK) == 0) {
                o = get_offset(&offset[file*MAX_O_TERMS],
                               &basfunc[file*MAX_O_TERMS], o_order,
                               nsamp[file], samp, det);
                get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                       &x, &y, badsamp[file][samp-1]);
                if (x > 0) {
                  b = get_background(&background[file*B_TERMS], samp, det);
                  im = ((rawdata[file][(samp-1)*ndet + (det-1)] - o - d)
                        /gain[gset[file]][det] - b) / skytran[file][samp-1];
/*                rms = im_rms[(y-1)*IM_X + (x-1)]; */
                  rms = sqrt(1.0/p_wt[det]) / skytran[file][samp-1];
/*                if (fabs((im - oldim[(y-1)*IM_X + (x-1)])
                         /rms/sqrt(oldwt[(y-1)*IM_X + (x-1)])) < deg_thresh) {
                                                                            */
                  if (fabs((im - oldim[(y-1)*IM_X + (x-1)])/rms) < deg_thresh) {
                    image[(y-1)*IM_X + (x-1)] += im*p_wt[det];
                    weight[(y-1)*IM_X + (x-1)] += 1.0*p_wt[det];
                  }
                  else {
                    (*deg_count)++;
                  }
                }
              }
            }
          }
        }
      }
     }
    }
    fprintf(stderr, "%d bolometer-frames deglitched\n", *deg_count);
    free (oldim);
    free (oldwt);

    sum0 = 0.0;
    sum1 = 0.0;
    avg_wt = avg_p_wt(p_wt, baddet, ndet);
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (weight[(y-1)*IM_X + (x-1)] > (0.1*avg_wt)) {
          image[(y-1)*IM_X + (x-1)] /= weight[(y-1)*IM_X + (x-1)];
          sum1 += image[(y-1)*IM_X + (x-1)];
          sum0 += 1.0;
        }
      }
    }
    *mean = sum1/sum0;
    if (rz_flag == 'y')
      *mean = 0.0;
    for (x = 1; x <= IM_X; x++) {
      for (y = 1; y <= IM_Y; y++) {
        if (weight[(y-1)*IM_X + (x-1)] > (0.1*avg_wt)) {
          image[(y-1)*IM_X + (x-1)] -= *mean;
          weight[(y-1)*IM_X + (x-1)] /= avg_wt;
        }
      }
    }
}

/*******************/

void calc_offsets(float *image, float *weight, float *skytran[],
                  float *rawdata[], float *gain[], int gset[], int baddet[],
                  char *badsamp[], float *offset[], float *basfunc[],
                  int o_order, float *background[], float *drift[],
                  int d_order, float *x0[], float *y0[], float *x1[],
                  float *y1[], int nfiles, long nsamp[], long ndet,
                  char *scanning[], char scanning_flag, char deep_flag,
                  float *rowavg[])

{
    int  file;
    long  samp, det;
    int  x, y;
    float  g, b, d, m, mw;
    int  term, row;
    double  da, sum0[MAX_O_TERMS], sum1[MAX_O_TERMS], mean;

    /* Calculate even for bad detectors */
    for (file = 0; file < nfiles; file++) {
      for (det = 1; det <= ndet; det++) {
        for (term = 0; term <= o_order; term++) {
          sum0[term] = 0.0;
          sum1[term] = 0.0;
        }
        g = gain[gset[file]][det];
        for (samp = 1; samp <= nsamp[file]; samp++) {
          if (scanning_flag || scanning[file][samp-1]) {
            if (badsamp[file][samp-1] == 0) {
              get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                     &x, &y, badsamp[file][samp-1]);
              b = get_background(&background[file*B_TERMS], samp, det);
              d = get_drift(&drift[file*MAX_O_TERMS],
                            &basfunc[file*MAX_O_TERMS], o_order, d_order,
                            nsamp[file], samp, det, deep_flag,
                            &rowavg[file*ARRAY_ROWS]);
              if (x > 0) {
                m = image[(y-1)*IM_X + (x-1)];
              }
	      else {
                m = 0.0;
              }
              mw = 1.0;
              /* Note:  Any non-uniform weighting will destroy orthogonality
                        of basis functions. */
/*            if (m > 0.0)
                mw = exp(-1.0*m/FLUX0); */
              da = rawdata[file][(samp-1)*ndet + (det-1)]
                  - g*skytran[file][samp-1]*m - g*b - d;
              for (term = 0; term <= o_order; term++) {
                sum1[term] += mw*da*basfunc[file*MAX_O_TERMS+term][samp-1];
                sum0[term] += mw*basfunc[file*MAX_O_TERMS+term][samp-1]
                               *basfunc[file*MAX_O_TERMS+term][samp-1];
              }
            }
          }
        }
        for (term = 0; term <= o_order; term++) {
          offset[file*MAX_O_TERMS+term][det-1] = sum1[term];
          if (sum0[term] > 0.1)
            offset[file*MAX_O_TERMS+term][det-1] /= sum0[term];
        }
      }
    }
}

/*******************/

void calc_drift(float *image, float *weight, float *skytran[],
                float *rawdata[], float *gain[], int gset[], int baddet[],
                char *badsamp[], float *offset[], float *basfunc[],
                int o_order, float *background[], float *drift[],
                int d_order, float *x0[], float *y0[], float *x1[],
                float *y1[], int nfiles, long nsamp[], long ndet,
                char *scanning[], char scanning_flag, float *rowavg[])

{
    int  file;
    long  samp, det, row, col;
    int  x, y;
    float  g, b, o;
    int  term;
    double  da, sum0[MAX_O_TERMS], sum1[MAX_O_TERMS], mean;
    float  list[MAX_LIST], ntile;
    int  listsize, k;

    /* IN FUTURE, TRY TO COMBINE LOOPS */

    for (file = 0; file < nfiles; file++) {
      for (row = 0; row < ARRAY_ROWS; row++) {
        for (term = 0; term < d_order; term++) {
          sum0[term] = 0.0;
          sum1[term] = 0.0;
        }
        for (col = 0; col < ARRAY_COLUMNS; col++) {
          det = row*ARRAY_COLUMNS + col + 1;
          if ((baddet[det] & BAD_MASK) == 0) {
            g = gain[gset[file]][det];
            for (samp = 1; samp <= nsamp[file]; samp++) {
              if (scanning_flag || scanning[file][samp-1]) {
                if (badsamp[file][samp-1] == 0) {
                  get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                         &x, &y, badsamp[file][samp-1]);
                  b = get_background(&background[file*B_TERMS], samp, det);
                  o = get_offset(&offset[file*MAX_O_TERMS],
                                 &basfunc[file*MAX_O_TERMS], o_order,
                                 nsamp[file], samp, det);
                  da = rawdata[file][(samp-1)*ndet + (det-1)] - g*b - o;
                  if (x > 0)
                    da -= g*skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
                  for (term = 0; term < d_order; term++) {
                    sum1[term] +=
                          da*basfunc[file*MAX_O_TERMS+term+o_order+1][samp-1];
                    sum0[term] +=
                          basfunc[file*MAX_O_TERMS+term+o_order+1][samp-1]
                          *basfunc[file*MAX_O_TERMS+term+o_order+1][samp-1];
                  }
                }
              }
            }
          }
        }
        for (term = 0; term < d_order; term++) {
          drift[file*MAX_O_TERMS+term][row] = sum1[term];
          if (sum0[term] > 0.1)
            drift[file*MAX_O_TERMS+term][row] /= sum0[term];
        }
      }
    }

    for (file = 0; file < nfiles; file++) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        for (row = 0; row < ARRAY_ROWS; row++) {
          listsize = 0;
          for (col = 0; col < ARRAY_COLUMNS; col++) {
            det = row*ARRAY_COLUMNS + col + 1;
            if ((baddet[det] & BAD_MASK) == 0) {
              g = gain[gset[file]][det];
              get_xy(det, samp, x0[file], y0[file], x1[file], y1[file], &x, &y,
                     badsamp[file][samp-1]);
              b = get_background(&background[file*B_TERMS], samp, det);
              o = get_offset(&offset[file*MAX_O_TERMS],
                             &basfunc[file*MAX_O_TERMS], o_order, nsamp[file],
                             samp, det);
              da = rawdata[file][(samp-1)*ndet + (det-1)] - g*b - o;
              if (x > 0)
                da -= g*skytran[file][samp-1]*image[(y-1)*IM_X + (x-1)];
              listsize++;
              list[listsize] = da;
            }
          }
          k = (listsize*5+5)/10;
          rowavg[file*ARRAY_ROWS+row][samp-1] =
                                       element_select(k, listsize, list);
          k = (listsize*5+10)/10;
          rowavg[file*ARRAY_ROWS+row][samp-1] +=
                                       element_select(k, listsize, list);
          rowavg[file*ARRAY_ROWS+row][samp-1] /= 2.0;
        }
      }
    }

}

/*******************/

void calc_gains(float *image, float *weight, float *skytran[], float *rawdata[],
                float *gain[], int gset[], int ngainsets, int baddet[],
                char *badsamp[], float *offset[], float *basfunc[], int o_order,
                float *background[], float *drift[], int d_order, float *x0[],
                float *y0[], float *x1[], float *y1[], int nfiles,
                long nsamp[], long ndet, char *scanning[], char scanning_flag,
                char deep_flag, float *rowavg[])

{
    int  file;
    long  samp, det, set;
    int  x, y;
    float  o, d, b, im;
    double  sum0, sum1, mean;

  /* Calculate even for bad detectors */
  for (set = 0; set < ngainsets; set++) {
    for (det = 1; det <= ndet; det++) {
      sum0 = 0.0;
      sum1 = 0.0;
      for (file = 0; file < nfiles; file++) {
       if (gset[file] == set) {
        for (samp = 1; samp <= nsamp[file]; samp++) {
          o = get_offset(&offset[file*MAX_O_TERMS], &basfunc[file*MAX_O_TERMS],
                         o_order, nsamp[file], samp, det);
          d = get_drift(&drift[file*MAX_O_TERMS], &basfunc[file*MAX_O_TERMS],
                        o_order, d_order, nsamp[file], samp, det, deep_flag,
                        &rowavg[file*ARRAY_ROWS]);
          if (scanning_flag || scanning[file][samp-1]) {
            if (badsamp[file][samp-1] == 0) {
              get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                     &x, &y, badsamp[file][samp-1]);
              b = get_background(&background[file*B_TERMS], samp, det);
              if (x > 0) {
                im = skytran[file][samp-1]*image[(y-1)*IM_X+(x-1)];
              }
	      else {
                im = 0.0;
              }
              sum1 += (rawdata[file][(samp-1)*ndet + (det-1)] -o-d)*(im + b);
              sum0 += (im + b)*(im + b);
            }
          }
        }
       }
      }
      gain[set][det] = sum1/sum0;
    }
  }
}

/*******************/

void norm_gains(float *gain[], int ngainsets, int baddet[], long ndet)

{
    long  det, set;
    float  list[MAX_LIST], ntile;
    int  listsize, k;

    for (set = 0; set < ngainsets; set++) {
      listsize = 0;
      for (det = 1; det <= ndet; det++) {
        if ((baddet[det] & BAD_MASK) == 0) {
          listsize++;
          list[listsize] = gain[set][det];
        }
      }
      k = (listsize*5+5)/10;
      ntile = element_select(k, listsize, list);
      k = (listsize*5+10)/10;
      ntile += element_select(k, listsize, list);
      ntile /= 2.0;     /* median value */
      for (det = 1; det <= ndet; det++) {
        gain[set][det] /= ntile;
      }
    }
}

/*******************/

float avg_p_wt(float p_wt[], int baddet[], long ndet)

{
    long  det;
    float  list[MAX_LIST], ntile;
    int  listsize, k;

    listsize = 0;
    for (det = 1; det <= ndet; det++) {
      if ((baddet[det] & BAD_MASK) == 0) {
        listsize++;
        list[listsize] = p_wt[det];
      }
    }
    k = (listsize*5+5)/10;
    ntile = element_select(k, listsize, list);
    k = (listsize*5+10)/10;
    ntile += element_select(k, listsize, list);
    ntile /= 2.0;

    return (ntile);
}

/*******************/

void calc_background(float *image, float *weight, float *skytran[],
                     float *rawdata[], float *gain[], int gset[], float p_wt[],
                     int baddet[], char *badsamp[], float *offset[],
                     float *basfunc[], int o_order, float *background[],
                     float *drift[], int d_order, float *x0[], float *y0[],
                     float *x1[], float *y1[], int nfiles, long nsamp[],
                     long ndet, char *scanning[], char scanning_flag,
                     char gradient_flag, char quadratic_flag, char deep_flag,
                     float *rowavg[])

{
    int  file;
    long  samp, det;
    int  x, y;
    float  o, d, g, m, dx, dy, da;
    float  **a, **b, *f;
    int  i, j;
    float  mw;

    /* Source space minimization now (9/30/2005) */
    /* Calculate even for bad frames */
    a = matrix(1, 4, 1, 4);
    b = matrix(1, 4, 1, 1);
    f = vector(1, 4);
    for (file = 0; file < nfiles; file++) {
      for (samp = 1; samp <= nsamp[file]; samp++) {
        for (i = 1; i <= 4; i++) {
          for (j = 1; j <= 4; j++) {
            a[i][j] = 0.0;
          }
          b[i][1] = 0.0;
        }
        for (det = 1; det <= ndet; det++) {
          if ((det % ARRAY_COLUMNS) == 1) {
            d = get_drift(&drift[file*MAX_O_TERMS], &basfunc[file*MAX_O_TERMS],
                          o_order, d_order, nsamp[file], samp, det, deep_flag,
                          &rowavg[file*ARRAY_ROWS]);
          }
          if ((baddet[det] & BAD_MASK) == 0) {
            g = gain[gset[file]][det];
            calc_pixel_xy(det, &dx, &dy);
            get_xy(det, samp, x0[file], y0[file], x1[file], y1[file],
                   &x, &y, badsamp[file][samp-1]);
            o = get_offset(&offset[file*MAX_O_TERMS],
                           &basfunc[file*MAX_O_TERMS], o_order, nsamp[file],
                           samp, det);
            if (x > 0) {
              m = image[(y-1)*IM_X + (x-1)];
            }
	    else {
              m = 0.0;
            }
            mw = 1.0*p_wt[det];
/*          if (m > 0.0)
              mw *= exp(-1.0*m/FLUX0); */
            da = rawdata[file][(samp-1)*ndet + (det-1)]/g
                       -skytran[file][samp-1]*m - o/g - d/g;
            f[1] = 1.0;
            f[2] = dx;
            f[3] = dy;
            f[4] = dx*dx;
            for (i = 1; i <= 4; i++) {
              for (j = 1; j <= 4; j++) {
                a[i][j] += mw*f[i]*f[j];
              }
              b[i][1] += mw*f[i]*da;
            }
          }
        }
        if (gradient_flag != 'y') {
          a[2][2] = 1.0;
          a[3][3] = 1.0;
          b[2][1] = 0.0;
          b[3][1] = 0.0;
          a[1][2] = 0.0;
          a[1][3] = 0.0;
          a[2][1] = 0.0;
          a[2][3] = 0.0;
          a[2][4] = 0.0;
          a[3][1] = 0.0;
          a[3][2] = 0.0;
          a[3][4] = 0.0;
          a[4][2] = 0.0;
          a[4][3] = 0.0;
        }
        if (quadratic_flag != 'y') {
          a[4][4] = 1.0;
          b[4][1] = 0.0;
          a[1][4] = 0.0;
          a[2][4] = 0.0;
          a[3][4] = 0.0;
          a[4][1] = 0.0;
          a[4][2] = 0.0;
          a[4][3] = 0.0;
        }

/* Reconsider next line when flux weighting is used */
        if (a[1][1] < 0.1) {
          if (samp > 1) {
            background[file*B_TERMS][samp-1] =
                                           background[file*B_TERMS][samp-2];
            background[file*B_TERMS+1][samp-1] =
                                           background[file*B_TERMS+1][samp-2];
            background[file*B_TERMS+2][samp-1] =
                                           background[file*B_TERMS+2][samp-2];
            background[file*B_TERMS+3][samp-1] =
                                           background[file*B_TERMS+3][samp-2];
          }
          else {
            background[file*B_TERMS][samp-1] = 0.0;
            background[file*B_TERMS+1][samp-1] = 0.0;
            background[file*B_TERMS+2][samp-1] = 0.0;
            background[file*B_TERMS+3][samp-1] = 0.0;
          }
        }
        else {
          gaussj(a, 4, b, 1);
          for (i = 1; i <= 4; i++)
            background[file*B_TERMS+(i-1)][samp-1] = b[i][1];
        }
      }
    }
    free_matrix(a, 1, 4, 1, 4);
    free_matrix(b, 1, 4, 1, 1);
    free_vector(f, 1, 4);
}

/*******************/

void zero_background(float *background[], int nfiles, long nsamp[],
                     char *badsamp[], float *azo[], float *azerr[], float *el[],
                     float *zao[], float *elerr[], float *rao[], float *deco[])

{
    int  file, i;
    long  samp;
    double  sum0, sum1, mean;
    double  sumx1, meanx;
    double  sumy1, meany;
    double  sumq1, meanq;
    float  off, offmin, offmax;

    for (file = 0; file < nfiles; file++) {
      /* set mean background terms to zero: */
      sum1 = 0.0;
      sum0 = 0.0;
      sumx1 = 0.0;
      sumy1 = 0.0;
      sumq1 = 0.0;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (badsamp[file][samp-1] == 0) {
          sum1 += background[file*B_TERMS][samp-1];
          sum0 += 1.0;
          sumx1 += background[file*B_TERMS+1][samp-1];
          sumy1 += background[file*B_TERMS+2][samp-1];
          sumq1 += background[file*B_TERMS+3][samp-1];
        }
      }
      mean = sum1/sum0;
      meanx = sumx1/sum0;
      meany = sumy1/sum0;
      meanq = sumq1/sum0;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        background[file*B_TERMS][samp-1] -= mean;
        background[file*B_TERMS+1][samp-1] -= meanx;
        background[file*B_TERMS+2][samp-1] -= meany;
        background[file*B_TERMS+3][samp-1] -= meanq;
      }

      /* get rid of azo correlated background: */
      sum1 = 0.0;
      sum0 = 0.0;
      offmin = 1.0e30;
      offmax = -1.0e30;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (badsamp[file][samp-1] == 0) {
          off = azo[file][samp-1]
                + azerr[file][samp-1]*cos(el[file][samp-1]*PI/180.0);
          sum1 += background[file*B_TERMS][samp-1]*off;
          sum0 += off*off;
          if (off > offmax) offmax = off;
          if (off < offmin) offmin = off;
        }
      }
      if ((offmin < -4.0) && (offmax > 4.0)) {
        mean = sum1/sum0;
        for (samp = 1; samp <= nsamp[file]; samp++) {
          off = azo[file][samp-1]
                + azerr[file][samp-1]*cos(el[file][samp-1]*PI/180.0);
          background[file*B_TERMS][samp-1] -= mean*off;
        }
      }

      /* get rid of zao correlated background: */
      sum1 = 0.0;
      sum0 = 0.0;
      offmin = 1.0e30;
      offmax = -1.0e30;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (badsamp[file][samp-1] == 0) {
          off = zao[file][samp-1] - elerr[file][samp-1];
          sum1 += background[file*B_TERMS][samp-1]*off;
          sum0 += off*off;
          if (off > offmax) offmax = off;
          if (off < offmin) offmin = off;
        }
      }
      if ((offmin < -4.0) && (offmax > 4.0)) {
        mean = sum1/sum0;
        for (samp = 1; samp <= nsamp[file]; samp++) {
          off = zao[file][samp-1] - elerr[file][samp-1];
          background[file*B_TERMS][samp-1] -= mean*off;
        }
      }

      /* get rid of rao correlated background: */
      /* Note not using azerr, elerr */
      sum1 = 0.0;
      sum0 = 0.0;
      offmin = 1.0e30;
      offmax = -1.0e30;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (badsamp[file][samp-1] == 0) {
          off = rao[file][samp-1];
          sum1 += background[file*B_TERMS][samp-1]*off;
          sum0 += off*off;
          if (off > offmax) offmax = off;
          if (off < offmin) offmin = off;
        }
      }
      if ((offmin < -4.0) && (offmax > 4.0)) {
        mean = sum1/sum0;
        for (samp = 1; samp <= nsamp[file]; samp++) {
          off = rao[file][samp-1];
          background[file*B_TERMS][samp-1] -= mean*off;
        }
      }

      /* get rid of deco correlated background: */
      /* Note not using azerr, elerr */
      sum1 = 0.0;
      sum0 = 0.0;
      offmin = 1.0e30;
      offmax = -1.0e30;
      for (samp = 1; samp <= nsamp[file]; samp++) {
        if (badsamp[file][samp-1] == 0) {
          off = deco[file][samp-1];
          sum1 += background[file*B_TERMS][samp-1]*off;
          sum0 += off*off;
          if (off > offmax) offmax = off;
          if (off < offmin) offmin = off;
        }
      }
      if ((offmin < -4.0) && (offmax > 4.0)) {
        mean = sum1/sum0;
        for (samp = 1; samp <= nsamp[file]; samp++) {
          off = deco[file][samp-1];
          background[file*B_TERMS][samp-1] -= mean*off;
        }
      }
    }
}

/*******************/

void block_average(float *rawdata, long ndet, long *nsamp)
 
{
    long  samp, newsamp, count, det;
    double  *sum1, sum0;
 
    sum1 = (double *) malloc(ndet*sizeof(double));
    for (det = 1; det <= ndet; det++) {
      sum1[det-1] = 0.0;
    }
    sum0 = 0.0;
    samp = 0;    /* Note using 0 to nsamp-1 in this case */
    newsamp = 0;
    count = 0;
    while (samp < *nsamp) {
      for (det = 1; det <= ndet; det++) {
        sum1[det-1] += rawdata[samp*ndet + (det-1)];
      }
      sum0 += 1.0;
      count++;
      if ((samp == (*nsamp -1)) || (count == BLOCK_SIZE)) {
        for (det = 1; det <= ndet; det++) {
          rawdata[newsamp*ndet + (det-1)] = sum1[det-1]/sum0;
        }
        for (det = 1; det <= ndet; det++) {
          sum1[det-1] = 0.0;
        }
        sum0 = 0.0;
        newsamp++;
        count = 0;
      }
      samp++;
    }
 
    free(sum1);
    *nsamp = newsamp;
}

/*******************/

void precess2000(float ra_1, float dec_1, float eq_1, float *ra_0, float *dec_0)

/* From Astronomical Almanac, p. B19 */

{
    float  T, M, N, am, dm;
    float  a, d;

    T = (eq_1 - 2000.0)/100.0;
    M = 1.2812323*T + 0.0003879*T*T + 0.0000101*T*T*T;
    M *= (PI/180.0);
    N = 0.5567530*T - 0.0001185*T*T + 0.0000116*T*T*T;
    N *= (PI/180.0);

    a = ra_1*15.0*PI/180.0;
    d = dec_1*PI/180.0;
    am = a - 0.5*(M + N*sin(a)*tan(d));
    dm = d - 0.5*N*cos(am);

    *ra_0 = -1.0*M - N*sin(am)*tan(dm);
    *ra_0 *= (180.0/PI/15.0);
    *ra_0 += ra_1;
    *dec_0 = -1.0*N*cos(am);
    *dec_0 *= (180.0/PI);
    *dec_0 += dec_1;
}

/*******************/

void printabort(char *message)

{
    fprintf(stderr, "ERROR:  %s\n", message);
    system("date >> sharcsolve_errlog.txt");
    system("echo \"ABORTING:  unrecoverable error\" >> sharcsolve_errlog.txt");
    exit (1);
}

/*******************/

void printerror(int status)

{
    if (status)
    {
       fits_report_error(stderr, status);
       system("date >> sharcsolve_errlog.txt");
       system("echo \"ABORTING:  FITS error\" >> sharcsolve_errlog.txt");
       exit (status);
    }
    return;
}
